#: Enumeration Types 

from enum import IntEnum


#: L23 Enums
class ReservedStatus(IntEnum):
    """Test resource reservation status"""
    RELEASED = 0
    """Test resource is released."""
    RESERVED_BY_YOU = 1
    """Test resource is reserved by you."""
    RESERVED_BY_OTHER = 2
    """Test resource is reserved by another user."""


class ReservedAction(IntEnum):
    """Reservation actions."""
    RELEASE = 0
    """Release the test resource."""
    RESERVE = 1
    """Reserve the test resource."""
    RELINQUISH = 2
    """Force release other's ownership of the test resource."""


class ChassisShutdownAction(IntEnum):
    """Chassis shutdown actions."""
    RESTART = 1
    """Restart after shutdown."""
    POWEROFF = 2
    """Keep power-off after shutdown."""


class OnOff(IntEnum):
    """On or Off."""
    OFF = 0
    """Off"""
    ON = 1
    """On"""


class RESTControlAction(IntEnum):
    """Controls of Valkyrie REST server."""
    START = 0
    """Start the REST server on the chassis."""
    STOP = 1
    """Stop the REST server on the chassis."""
    RESTART = 2
    """Restart the REST server on the chassis."""


class ServiceStatus(IntEnum):
    """Service status."""
    SERVICE_OFF = 0
    """Service is off."""
    SERVICE_ON = 1
    """Service is on."""


class ChassisSessionType(IntEnum):
    """The type of the session between the client and the chassis."""
    MANAGER = 1
    """Connection on port 22606 running binary command (used by ValkyrieManager & VulcanManager)."""
    SCRIPT = 2
    """Connection on port 22611, running CLI command."""


class TimeSyncMode(IntEnum):
    """Module's time sync mode."""
    CHASSIS = 0
    """Module syncs to chassis's clock."""
    EXTERNAL = 1
    """Module syncs to an external clock."""
    MODULE = 2
    """Module syncs to its own clock."""


class MediaCFPState(IntEnum):
    """Modules' CFP state"""
    NOTCFP = 0
    """This is not a CFP-based test module."""
    NOTPRESENT = 1
    """No transceiver, the CFP cage is empty"""
    NOTFLEXIBLE = 2
    """Transceiver present, supporting a fixed speed and port-count"""
    FLEXIBLE = 3
    """Transceiver present, supporting flexible speed and port-count"""


class MediaCFPType(IntEnum):
    """Module's Media CFP Type. What kind of transceiver, may have multiple ports"""
    CFP_UNKNOWN = 0
    """CFP unknown"""
    CFP_INVALID = 1
    """CFP 100G not supported"""
    CFP_1X40_SR4 = 2
    """CFP 40G SR4 850 nm"""
    CFP_2X40_SR4 = 3
    """CFP 40G SR4 850 nm"""
    CFP_1X40_LR4 = 4
    """CFP 40G LR4 1310 nm"""
    CFP_1X100_LR4 = 5
    """CFP 100G LR4 1310 nm"""
    CFP_1X100_SR10 = 6
    """CFP 100G SR10 850 nm / CFP 40G SR4 850 nm"""
    CFP_4X10_SFPP = 7
    """"""
    QSFP_EMPTY = 8
    """QSFP empty"""
    QSFP_UNKNOWN = 9
    """QSFP unknown"""
    QSFP_SR = 10
    """QSFP SR 850 nm"""
    QSFP_LR = 11
    """QSFP LR 1550 nm"""
    QSFP_ER = 12
    """QSFP ER 1550 nm"""
    CXP_UNKNOWN = 13
    """CXP unknown SR10"""
    CXP_SR10 = 14
    """CXP 100G SR10 850 nm / CXP 40G SR4 850 nm"""
    CXP_AOC = 15
    """CXP 100G SR10 AOC / CXP 40G SR4 AOC"""
    CFP4_UNKNOWN = 16
    """CFP4 unknown"""
    CFP4_1X100_LR4 = 17
    """CFP4 100G LR4"""
    CFP4_1X100_ER4 = 18
    """CFP4 100G ER4"""
    CFP4_1X100_SR4 = 19
    """CFP4 100G SR4"""
    QSFP28_CR4 = 20
    """QSFP28 CR4"""
    QSFP28_SR4 = 21
    """QSFP28 SR4"""
    QSFP28_LR4 = 22
    """QSFP28 LR4"""
    QSFP28_ER4 = 23
    """QSFP28 ER4"""
    QSFP28_AOC = 24
    """QSFP28 AOC"""
    QSFP28_ACC = 25
    """QSFP28 ACC"""
    QSFP28_UNKNOWN = 26
    """QSFP28 unknown"""
    QSFP28P_SR4 = 27
    """QSFP+ SR4 850 nm"""
    QSFP28P_LR4 = 28
    """QSFP+ LR4 1310 nm"""
    QSFP28P_ER4 = 29
    """QSFP+ ER4 1500 nm"""
    QSFP28P_CR4 = 30
    """QSFP+ CR4 Copper"""
    QSFP28P_ISM4 = 31
    """QSFP+ iSM4 1310 nm"""
    QSFP28P_UNKNOWN = 32
    """QSFP+ unknown"""
    TCVR_UNKNOWN = 33
    """Transceiver unknown"""
    QSFP28_CWDM4 = 34
    """QSFP28 CWDM4"""
    QSFP28_CWDM4FEC = 35
    """QSFP28 CWDM4 FEC"""
    QSFP28_PSM4 = 36
    """QSFP28 PSM4"""


class SMAInputFunction(IntEnum):
    """SMA input function"""
    NOTUSED = 0
    """SMA input not used"""
    TX2MHZ = 1
    """TX Clock Ref. 2.048 MHz"""
    TX10MHZ = 2
    """TX Clock Ref. 10.0 MHz"""


class SMAOutputFunction(IntEnum):
    """SMA output function"""
    DISABLED = 0
    """Disabled"""
    PASSTHROUGH = 1
    """Pass-through"""
    P0SOF = 2
    """Port 0 Start-of-Frame Pulse"""
    P1SOF = 3
    """Port 1 Start-of-Frame Pulse"""
    REF2MHZ = 4
    """TX Clock (nom. 2.048 MHz)"""
    REF10MHZ = 5
    """TX Clock (nom. 10.0 MHz)"""
    REF125MHZ = 6
    """TX Clock (nom. 125 MHz)"""
    REF156MHZ = 7
    """TX Clock (nom. 156.25 MHz)"""
    P0RXCLK = 8
    """Port 0 RX Clock (nom. 10.0 MHz)"""
    P1RXCLK = 9
    """Port 1 RX Clock (nom. 10.0 MHz)"""
    P0RXCLK2MHZ = 10
    """Port 0 RX Clock (nom. 2.048 MHz)"""
    P1RXCLK2MHZ = 11
    """Port 1 RX Clock (nom. 2.048 MHz)"""
    TS_PPS = 12
    """Timing Source (Pulse-Per-Second)"""


class SMAStatus(IntEnum):
    """SMA status"""
    OK = 0
    """Status OK"""
    NO_VALID_SIGNAL = 1
    """No valid signal"""


class HasDemo(IntEnum):
    """Demo or not"""
    NON_DEMO = 0
    """Not a demo unit"""
    DEMO = 1
    """Demo unit"""


class IsValid(IntEnum):
    """Whether it is valid"""
    INVALID = 0
    """Invalid"""
    VALID = 1
    """Valid"""


class IsPermanent(IntEnum):
    """Whether it is permanent"""
    NON_PERMANENT = 0
    """Not permanent"""
    PERMANENT = 1
    """Permanent"""


class YesNo(IntEnum):
    NO = 0
    """No"""
    YES = 1
    """Yes"""


class UpdateState(IntEnum):
    """License update state"""
    NONE = 0
    """None"""
    UPDATING = 1
    """Updating"""
    UPDATE_SUCCESS = 2
    """Update successful"""
    UPDATE_FAIL = 3
    """Update failed"""


class IsOnline(IntEnum):
    """Chassis online or not"""
    OFFLINE = 0
    """Chassis offline"""
    ONLINE = 1
    """Chassis online"""


class PortSpeedMode(IntEnum):
    """Port speed mode"""
    AUTO = 0
    """Auto negotiate with peer"""
    F10M = 1
    """Forced to 10 Mbit/s"""
    F100M = 2
    """Forced to 100 Mbit/s"""
    F1G = 3
    """Forced to 1 Gbit/s"""
    F10G = 4
    """Forced to 10 Gbit/s"""
    F40G = 5
    """Forced to 40 Gbit/s"""
    F100G = 6
    """Forced to 100 Gbit/s"""
    F10MHDX = 7
    """Forced to 10 Mbit/s half-duplex"""
    F100MHDX = 8
    """Forced to 100 Mbit/s half-duplex"""
    F10M100M = 9
    """Auto negotiate to 10/100 Mbit/s"""
    F100M1G = 10
    """Auto negotiate to 100 Mbit/s or 1 Gbit/s"""
    F100M1G10G = 11
    """Auto negotiate to 100 Mbit/s or 1 Gbit/s or 10 Gbit/s"""
    F2500M = 12
    """Forced to 2500 Mbit/s"""
    F5G = 13
    """Forced to 5 Gbit/s"""
    F100M1G2500M = 14
    """Auto negotiate to 100 Mbit/s or 1 Gbit/s or 2500 Mbit/s"""
    F25G = 15
    """Forced to 25 Gbit/s"""
    F50G = 16
    """Forced to 50 Gbit/s"""
    F200G = 17
    """Forced to 200 Gbit/s"""
    F400G = 18
    """Forced to 400 Gbit/s"""
    F800G = 19
    """Forced to 800 Gbit/s"""
    F1600G = 20
    """Forced to 1600 Gbit/s"""
    UNKNOWN = 255
    """Speed mode unknown"""


class SyncStatus(IntEnum):
    """Port sync status"""
    NO_SYNC = 0
    """Not sync"""
    IN_SYNC = 1
    """In sync with peer"""


class LoopbackMode(IntEnum):
    """Port loopback mode"""
    NONE = 0
    """Off"""
    L1RX2TX = 1
    """L1 RX-to-TX"""
    L2RX2TX = 2
    """L2 RX-to-TX"""
    L3RX2TX = 3
    """L3 RX-to-TX"""
    TXON2RX = 4
    """TX(on)-to-RX"""
    TXOFF2RX = 5
    """TX(off)-to-RX"""
    PORT2PORT = 6
    """Port-to-Port"""


class TrafficOnOff(IntEnum):
    """Traffic status"""
    OFF = 0
    """Traffic off"""
    ON = 1
    """Traffic on"""
    PREPARING = 2
    """Traffic preparing"""


class StartOrStop(IntEnum):
    STOP = 0
    """Stop the action"""
    START = 1
    """Start the action"""


class LatencyMode(IntEnum):
    """Latency measurement mode"""
    LAST2LAST = 0
    """Last-to-Last"""
    FIRST2LAST = 1
    """First-to-Last"""
    LAST2FIRST = 2
    """Last-to-First"""
    FIRST2FIRST = 3
    """First-to-First"""


class SourceType(IntEnum):
    TXIFG = 0
    """TX IFG"""
    TXLEN = 1
    """TX Length"""
    RXIFG = 2
    """RX IFG"""
    RXLEN = 3
    """RX Length"""
    RXLAT = 4
    """RX Latency"""
    RXJIT = 5
    """RX Jitter"""


class PacketDetailSelection(IntEnum):
    ALL = 0
    """All"""
    TPLD = 1
    """Based on Test Payload ID"""
    FILTER = 2
    """Based on filter"""


class OnOffWithSuppress(IntEnum):
    """Stream status"""
    OFF = 0
    """Off"""
    ON = 1
    """On"""
    SUPPRESS = 2
    """Suppressed"""


class ProtocolOption(IntEnum):
    """Protocol header options"""
    ETHERNET = 1
    """Ethernet II"""
    VLAN = 2
    """VLAN"""
    ARP = 3
    """Address Resolution Protocol"""
    IP = 4
    """IPv4"""
    IPV6 = 5
    """IPv6"""
    UDP = 6
    """User Datagram Protocol (w/o checksum)"""
    TCP = 7
    """Transmission Control Protocol (w/o checksum)"""
    LLC = 8
    """Logic Link Control"""
    SNAP = 9
    """Subnetwork Access Protocol"""
    GTP = 10
    """GPRS Tunnelling Protocol"""
    ICMP = 11
    """Internet Control Message Protocol"""
    RTP = 12
    """Real-time Transport Protocol"""
    RTCP = 13
    """RTP Control Protocol"""
    STP = 14
    """Spanning Tree Protocol"""
    SCTP = 15
    """Stream Control Transmission Protocol"""
    MACCTRL = 16
    """MAC Control"""
    MPLS = 17
    """Multiprotocol Label Switching"""
    PBBTAG = 18
    """Provider Backbone Bridge tag"""
    FCOE = 19
    """Fibre Channel over Ethernet"""
    FC = 20
    """Fibre Channel"""
    FCOETAIL = 21
    """Fibre Channel over Ethernet (tail)"""
    IGMPV3L0 = 22
    """IGMPv3 Membership Query L=0"""
    IGMPV3L1 = 23
    """IGMPv3 Membership Query L=1"""
    UDPCHECK = 24
    """User Datagram Protocol (w/ checksum)"""
    IGMPV2 = 25
    """Internet Group Management Protocol v2"""
    MPLS_TP_OAM = 26
    """MPLS-TP, OAM Header"""
    GRE_NOCHECK = 27
    """Generic Routing Encapsulation (w/o checksum)"""
    GRE_CHECK = 28
    """Generic Routing Encapsulation (w/ checksum)"""
    TCPCHECK = 29
    """Transmission Control Protocol (w/ checksum)"""
    GTPV1L0 = 30
    """GTPv1 (no options), GPRS Tunneling Protocol v1"""
    GTPV1L1 = 31
    """GTPv1 (w/ options), GPRS Tunneling Protocol v1"""
    GTPV2L0 = 32
    """GTPv2 (no options), GPRS Tunneling Protocol v2"""
    GTPV2L1 = 33
    """GTPv2 (w/ options), GPRS Tunneling Protocol v2"""
    IGMPV1 = 34
    """Internet Group Management Protocol v1"""
    PWETHCTRL = 35
    """PW Ethernet Control Word"""
    VXLAN = 36
    """Virtual eXtensible LAN"""
    ETHERNET_8023 = 37
    """Ethernet 802.3"""
    NVGRE = 38
    """Generic Routing Encapsulation (Network Virtualization)"""
    DHCPV4 = 39
    """Dynamic Host Configuration Protocol (IPv4)"""
    GENEVE = 40
    """Generic Network Virtualization Encapsulation"""
    XENA_TPLD = 41
    XENA_TPLD_PI = 42
    XENA_MICROTPLD = 43
    ETHERNET_FCS = 44
    MACCTRLPFC = 45
    """MAC Control for PFC"""
    ECPRI = 46
    """Enhanced Common Public Radio Interface"""
    ROE = 47
    """Radio over Ethernet"""
    ETHERTYPE = 48
    """EtherType"""
    
    # Generat RAW form 1...64 bytes
    _ignore_ = 'ProtocolOption i'
    ProtocolOption = vars()
    for i in range(1, 65):
        ProtocolOption['RAW_%d' % i] = 256 - i # type: ignore

class ModifierAction(IntEnum):
    """Modifier action mode"""
    INC = 0
    """Incrementing"""
    DEC = 1
    """Decrementing"""
    RANDOM = 2
    """Random"""


class LengthType(IntEnum):
    """Packet length type"""
    FIXED = 0
    """Fixed"""
    INCREMENTING = 1
    """Incrementing"""
    BUTTERFLY = 2
    """Butterfly"""
    RANDOM = 3
    """Random"""
    MIX = 4
    """Mix"""


class PayloadType(IntEnum):
    """Packet payload type"""
    PATTERN = 0
    """Pattern"""
    INCREMENTING = 1
    """Incrementing"""
    PRBS = 2
    """PRBS"""
    RANDOM = 3
    """Random"""


class MDIXMode(IntEnum):
    """MDIX mode"""
    AUTO = 0
    """Auto"""
    MDI = 1
    """MDI"""
    MDIX = 2
    """MDIX"""


class LengthCheckType(IntEnum):
    AT_MOST = 0
    """At Most"""
    AT_LEAST = 1
    """At Least"""


class StartTrigger(IntEnum):
    """Capture start trigger"""
    ON = 0
    """From Traffic On"""
    FCSERR = 1
    """From FCS Error"""
    FILTER = 2
    """From Filter"""
    PLDERR = 3
    """From Payload Error"""


class StopTrigger(IntEnum):
    """Capture stop trigger"""
    FULL = 0
    """Until Capture Buffer Full"""
    FCSERR = 1
    """Until Receiving FCS Error"""
    FILTER = 2
    """Until Filter Matched"""
    PLDERR = 3
    """Until Receiving Payload Error"""
    USERSTOP = 4
    """Until User Stops"""


class PacketType(IntEnum):
    """Type of Packet to Keep in Capture Buffer"""
    ALL = 0
    """All Packets"""
    FCSERR = 1
    """With FCS Errors"""
    NOTPLD = 2
    """Without Test Payload"""
    TPLD = 3
    """With Test Payload"""
    FILTER = 4
    """Filtered Packets"""
    PLDERR = 5
    """With Payload Errors"""


class InjectErrorType(IntEnum):
    """Lane Injection Error Type"""
    HEADERERROR = 1
    """Header Error"""
    ALIGNERROR = 2
    """Alignment Error"""
    BIP8ERROR = 3
    """BIP8 Error"""


class HeaderLockStatus(IntEnum):
    """Physical Lane Header Lock Status"""
    HEADEROFF = 0
    """Header Lock Off"""
    HEADERON = 1
    """Header Lock On"""
    HEADEROFFUNSTABLE = 2
    """Header Lock Off and Unstable"""
    HEADERONUNSTABLE = 3
    """Header Lock On but Unstable"""


class AlignLockStatus(IntEnum):
    """Physical Lane Alignment Lock Status"""
    ALIGNOFF = 0
    """Alignment Lock Off"""
    ALIGNON = 1
    """Alignment Lock On"""
    ALIGNOFFUNSTABLE = 2
    """Alignment Lock Off and Unstable"""
    ALIGNONUNSTABLE = 3
    """Alignment Lock On but Unstable"""


class PRBSLockStatus(IntEnum):
    """Physical Lane PRBS Lock Status"""
    PRBSOFF = 0
    """PRBS Lock Off"""
    PRBSON = 1
    """PRBS Lock On"""
    PRBSOFFUNSTABLE = 2
    """PRBS Lock Off and Unstable"""
    PRBSONUNSTABLE = 3
    """PRBS Lock On but Unstable"""


class MulticastOperation(IntEnum):
    """IGMPv2 Request Type"""
    OFF = 0
    """Off"""
    ON = 1
    """On"""
    JOIN = 2
    """Join"""
    LEAVE = 3
    """Leave"""


class MulticastExtOperation(IntEnum):
    """IGMPv2/v3Request Type"""
    OFF = 0
    """Off"""
    ON = 1
    """On"""
    JOIN = 2
    """Join"""
    LEAVE = 3
    """Leave"""
    INCLUDE = 4
    """Include"""
    EXCLUDE = 5
    """EXclude"""
    LEAVE_TO_ALL = 6
    """Leave To All"""
    GENERAL_QUERY = 7
    """General Query"""
    GROUP_QUERY = 8
    """Group Query"""


class IGMPVersion(IntEnum):
    """IGMP Version"""
    IGMPV2 = 0
    """IGMP Version 2"""
    IGMPV3 = 1
    """IGMP Version 3"""


class TXMode(IntEnum):
    """Port TX Mode"""
    NORMAL = 0
    """Normal"""
    STRICTUNIFORM = 1
    """Strict Uniform"""
    SEQUENTIAL = 2
    """Sequential"""
    BURST = 3
    """Burst"""


class PayloadMode(IntEnum):
    """Payload Mode"""
    NORMAL = 0
    """Normal"""
    EXTPL = 1
    """Extend Payload"""
    CDF = 2
    """Custom Data Field"""


class BRRMode(IntEnum):
    """BRR Mode"""
    SLAVE = 0
    """Slave Mode"""
    MASTER = 1
    """Master Mode"""


class TXClockSource(IntEnum):
    """TX Clock Source"""
    MODULELOCALCLOCK = 0
    """Module Local Clock"""
    SMAINPUT = 1
    """SMA Input"""
    P0RXCLK = 2
    """Port 0 RX Clock"""
    P1RXCLK = 3
    """Port 1 RX Clock"""
    P2RXCLK = 4
    """Port 2 RX Clock"""
    P3RXCLK = 5
    """Port 3 RX Clock"""
    P4RXCLK = 6
    """Port 4 RX Clock"""
    P5RXCLK = 7
    """Port 5 RX Clock"""
    P6RXCLK = 8
    """Port 6 RX Clock"""
    P7RXCLK = 9
    """Port 7 RX Clock"""


class TXClockStatus(IntEnum):
    """TX Clock Status"""
    OK = 0
    """Clock OK"""
    NOVALIDTXCLK = 1
    """No Valid TX Clock"""


class LoopBandwidth(IntEnum):
    """Loop Bandwidth"""
    BW103HZ = 1
    """Loop Bandwidth = 103 Hz"""
    BW207HZ = 2
    """Loop Bandwidth = 207 Hz"""
    BW416HZ = 3
    """Loop Bandwidth = 416 Hz"""
    BW1683HZ = 4
    """Loop Bandwidth = 1683 Hz"""
    BW7019HZ = 5
    """Loop Bandwidth = 7019 Hz"""


class MediaConfigurationType(IntEnum):
    """Module Media Configuration Type"""
    CFP4 = 0
    """CFP4"""
    QSFP28 = 1
    """QSFP28 (NRZ)"""
    CXP = 2
    """CXP"""
    SFP28 = 3
    """SFP28"""
    QSFP56 = 4
    """QSFP56 (PAM4)"""
    QSFP_DD = 5
    """QSFP-DD (PAM4)"""
    SFP56 = 6
    """SFP56"""
    SFP_DD = 7
    """SFP-DD"""
    SFP112 = 8
    """SFP112"""
    QSFP_DD_NRZ = 9
    """QSFP-DD (NRZ)"""
    QSFP28_PAM4 = 10
    """QSFP28 (PAM4)"""
    CFP = 99
    """CFP"""
    BASE_T1 = 100
    """BASE-T1"""
    BASE_T1S = 101
    """BASE-T1S"""
    QSFPDD800 = 110
    """QSFP-DD800"""
    QSFP112 = 111
    """QSFP112"""
    OSFP800 = 112
    """OSFP800"""
    UNKNOWN = 255


class TXHState(IntEnum):
    """Recent Change in EEE State on TX"""
    TXH_NA = 0
    """Leaving or Going Into"""
    TXH_X = 1
    """No Activity"""


class RXHState(IntEnum):
    """Recent Change in EEE State on RX"""
    RXH_NA = 0
    """Leaving or Going Into"""
    RXH_X = 1
    """No Activity"""


class TXCState(IntEnum):
    """TX EEE State"""
    TXC_ACTIVE = 0
    """Active"""
    TXC_LPI = 1
    """Low Power"""


class RXCState(IntEnum):
    """RX EEE State"""
    RXC_ACTIVE = 0
    """Active"""
    RXC_LPI = 1
    """Low Power"""


class LinkState(IntEnum):
    """Low Power Mode Link State"""
    LINK_DOWN = 0
    """Link Down"""
    LINK_UP = 1
    """Link Up"""


class FaultSignaling(IntEnum):
    """Fault Signaling Behavior"""
    NORMAL = 0
    """Normal"""
    FORCE_LOCAL = 1
    """Forced to Local"""
    FORCE_REMOTE = 2
    """Forced to Remote"""
    DISABLED = 3
    """Disabled"""


class LocalFaultStatus(IntEnum):
    """Local Fault Status"""
    OK = 0
    """OK"""
    LOCAL_FAULT = 1
    """Local Fault"""


class RemoteFaultStatus(IntEnum):
    """Remote Fault Status"""
    OK = 0
    """OK"""
    REMOTE_FAULT = 1
    """Remote Fault"""


class TPLDMode(IntEnum):
    """Test Payload Mode"""
    NORMAL = 0
    """Normal"""
    MICRO = 1
    """Micro"""


class SerdesStatus(IntEnum):
    """Serdes Status"""
    STOPPED = 0
    """Stopped"""
    STARTED = 1
    """Started"""
    INITIALIZING = 2
    """Initializing"""
    PLOTTING = 3
    """Plotting"""


class FECMode(IntEnum):
    """FEX Mode"""
    OFF = 0
    """Off"""
    ON = 1
    """On"""
    RS_FEC = 2
    """RS FEC"""
    FC_FEC = 3
    """Firecode FEC"""
    RS_FEC_KR = 4
    """RS FEC KR"""
    RS_FEC_KP = 5
    """RS FEC KP"""


class PRBSInsertedType(IntEnum):
    """PRBS Type"""
    CAUI_VIRTUAL = 0
    """CAUI Virtual"""
    PHY_LINE = 1
    """PHY Line"""
    PHY_HOST = 2
    """PHY Host"""
    TCVR = 3
    """Tranceiver"""


class PRBSPolynomial(IntEnum):
    """PRBS Polynomial"""
    PRBS7 = 0
    """PRBS-7"""
    PRBS9 = 1
    """PRBS-9"""
    PRBS11 = 2
    """PRBS-11"""
    PRBS15 = 3
    """PRBS-15"""
    PRBS23 = 4
    """PRBS-23"""
    PRBS31 = 5
    """PRBS-31"""
    PRBS58 = 6
    """PRBS-58"""
    PRBS49 = 7
    """PRBS-49"""
    PRBS10 = 8
    """PRBS-10"""
    PRBS20 = 9
    """PRBS-20"""
    PRBS13 = 10
    """PRBS-13"""


class PRBSInvertState(IntEnum):
    """PRBS Invert State"""
    NON_INVERTED = 0
    """Non-inverted"""
    INVERTED = 1
    """Inverted"""


class PRBSStatisticsMode(IntEnum):
    """PRBS Statistics Mode"""
    ACCUMULATIVE = 0
    """Accumulative"""
    PERSECOND = 1
    """Per Second"""


class AutoNegMode(IntEnum):
    """Auto Neg Mode"""
    ANEG_OFF = 0
    """Auto Neg Off"""
    ANEG_ON = 1
    """Auto Neg On"""


class AutoNegTecAbility(IntEnum):
    """Auto Neg Technical Abilities"""
    DEFAULT_TECH_MODE = 0
    """Default Tech Mode"""
    IEEE_10G_KR = 4
    """IEEE 10G KR"""
    IEEE_40G_CR4 = 16
    """IEEE 40G CR4"""
    IEEE_100G_KR4 = 128
    """IEEE 100G KR4"""
    IEEE_100G_CR4 = 256
    """IEEE 100G CR4"""
    IEEE_25GBASE_CRS_KRS = 512
    """IEEE 25GBASE CRS KRS"""
    IEEE_25GBASE_CR_KR = 1024
    """IEEE 25GBASE CR KR"""
    IEEE_50GBASE_CR_KR = 8192
    """IEEE 50GBASE CR KR"""
    IEEE_100GBASE_CR2_KR2 = 16384
    """IEEE 100GBASE CR2 KR2"""
    IEEE_200GBASE_CR4_KR4 = 32768
    """IEEE 200GBASE CR4 KR4"""
    EC_25GBASE_KR1 = 16777216
    """EC 25GBASE KR1"""
    EC_25GBASE_CR1 = 33554432
    """EC 25GBASE CR1"""
    EC_50GBASE_KR2 = 67108864
    """EC 50GBASE KR2"""
    EC_50GBASE_CR2 = 134217728
    """EC 50GBASE CR2"""
    EC_400GGBASE_KR8 = 268435456
    """EC 400GGBASE KR8"""
    EC_50G_CR1_KR1 = 503
    """EC 50G CR1 KR1"""
    BAM_50G_CR1_KR1 = 504
    """BAM 50G CR1 KR1"""
    BAM_50G_CR2_KR2 = 505
    """BAM 50G CR2 KR2"""
    BAM_100G_CR2_KR2 = 1002
    """BAM 100G CR2 KR2"""
    BAM_100G_CR4_KR4 = 1003
    """BAM 100G CR4 KR4"""
    BAM_200G_CR2_KR2 = 2002
    """BAM 200G CR2 KR2"""
    BAM_400G_CR8_KR8 = 4001
    """BAM 400G CR8 KR8"""


class AutoNegFECOption(IntEnum):
    """Auto Neg FEC Options"""
    DEFAULT_FEC = 0
    """Default FEC"""
    NO_FEC = 1
    """No FEC"""
    FCFEC = 2
    """Firecode FEC"""
    RSFEC_CL91 = 4
    """RS FEC Cl91"""
    RS528 = 256
    """RS 528"""
    RS544 = 512
    """RS 544"""
    RS272 = 1024
    """RS 272"""


class PauseMode(IntEnum):
    """Pause Mode"""
    NO_PAUSE = 0
    """No Pause"""
    SYM_PAUSE = 1
    """Symmetric Pause"""
    ASYM_PAUSE = 2
    """Asymmetric Pause"""


class AutoNegFECType(IntEnum):
    """Auto Neg FEC Type"""
    PENDING = 0
    """Pending"""
    NOFEC = 1
    """No FEC"""
    RS_FEC = 513
    """RS FEC"""
    FC_FEC = 257
    """Firecode FEC"""


class AutoNegStatus(IntEnum):
    """Auto Neg Status"""
    UNKNOWN = 0
    """Unknown"""
    ENABLE = 1
    """Enabled"""
    TRANSMIT_DISABLE = 2
    """Transmit Disabled"""
    ABILITY_DETECT = 3
    """Ability Detected"""
    ACKNOWLEDGE_DETECT = 4
    """Acknowledge Detected"""
    COMPLETE_ACKNOWLEDGE = 5
    """Complete Acknowledge"""
    NEXT_PAGE_WAIT = 6
    """Next Page Wait"""
    AN_GOOD_CHECK = 7
    """AN Good Check"""
    AN_GOOD = 8
    """AN Good"""


class AutoNegFECStatus(IntEnum):
    """Auto Neg FEC Status"""
    DEFAULT_FEC = 0
    """Default FEC"""
    NO_FEC = 1
    """No FEC"""
    FC_FEC = 2
    """Firecode FEC"""
    RSFEC_CL91 = 4
    """RS FEC Cl91"""
    RS528 = 256
    """RS 528"""
    RS544 = 512
    """RS 544"""
    RS272 = 1024
    """RS 272"""


class LinkTrainingMode(IntEnum):
    """Link Training Mode"""
    AUTO = 0
    """Auto"""
    FORCE_ENABLE = 1
    """Forced Enable"""


class PAM4FrameSize(IntEnum):
    """PAM4 Frame Size"""
    N16K_FRAME = 0
    N4K_FRAME = 1


class LinkTrainingInitCondition(IntEnum):
    """Link Training Intialization Condition"""
    NO_INIT = 0
    """No Intialization"""
    INIT_ENABLED = 1
    """Intialization Enabled"""


class NRZPreset(IntEnum):
    """Link Training NRZ Preset"""
    NRZ_NO_PRESET = 0
    """NRZ without Preset"""
    NRZ_WITH_PRESET = 1
    """NRZ with Preset"""


class TimeoutMode(IntEnum):
    """Link Training Timeout Mode"""
    DEFAULT_TIMEOUT = 0
    """Default Timeout"""
    TIMEOUT_DISABLED = 255
    """Timeout Disabled"""


class LinkTrainingStatusMode(IntEnum):
    """Link Training Status Mode"""
    DISABLED = 0
    """Disabled"""
    ENABLED = 1
    """Enabled"""


class LinkTrainingStatus(IntEnum):
    """Link Training Status"""
    NOT_TRAINED = 0
    """Not Trained"""
    TRAINED = 1
    """Trained"""


class LinkTrainingFailureType(IntEnum):
    """Link Training Failure Type"""
    NO_FAILURE = 0
    """No Failure"""
    FRAME_LOCK_ERROR = 1
    """Frame Lock Error"""
    SNR_BELOW_THRESHOLD = 2
    """SNR Below Threshold"""
    TIME_OUT_FAILURE = 3
    """Timeout Failure"""


class MulticastHeaderFormat(IntEnum):
    """Additional Header to IGMPv2/v3 Packets"""
    NOHDR = 0
    """No Header"""
    VLAN = 1
    """VLAN"""


class PFCMode(IntEnum):
    """Priority Flow Control (PFC) Mode"""
    VLAN_PCP = 128
    """VLAN PCP"""


class PRBSOnOff(IntEnum):
    """PRBS Status"""
    PRBSOFF = 0
    """PRBS Off"""
    PRBSON = 1
    """PRBS On"""


class ErrorOnOff(IntEnum):
    """PRBS Error Injection Status"""
    ERRORSOFF = 0
    """PRBS Error Injection Off"""
    ERRORSON = 1
    """PRBS Error Injection On"""


class PRBSPattern(IntEnum):
    """PRBS Pattern"""
    PRBS7 = 0
    """PRBS-7"""
    PRBS9 = 1
    """PRBS-9"""
    PRBS11 = 2
    """PRBS-11"""
    PRBS15 = 3
    """PRBS-15"""
    PRBS23 = 4
    """PRBS-23"""
    PRBS31 = 5
    """PRBS-31"""


class PHYSignalStatus(IntEnum):
    """PHY Signal Status"""
    NO_SIGNAL = 0
    """No Signal"""
    NO_CDRLOCK = 2
    """No CDR Lock"""
    LOCKED = 3
    """Locked"""


class OnOffDefault(IntEnum):
    """On Off Default Status"""
    OFF = 0
    """Off"""
    ON = 1
    """On"""
    DEFAULT = 2
    """Default"""


class TimeKeeperLicenseFileState(IntEnum):
    """TimeKeeper License File State"""
    NA = 0
    """Not Available"""
    INV = 1
    """Invalid"""
    VALID = 2
    """Valid"""


class TimeKeeperLicenseType(IntEnum):
    """TimeKeeper License Type"""
    UNDEF = 0
    """Undefined"""
    CLIENT = 1
    """TimeKeeper Client"""
    SERVER = 2
    """TimeKeeper Server"""


class TimeKeeperLicenseError(IntEnum):
    """TimeKeeper License Error"""
    NO_LICENSE_ERROR = 0
    """No License Error"""
    INVALID_SERIALNO = 1
    """Invalid Serial Number"""
    INVALID_CHASSISTYPE = 2
    """Invalid Chassis Type"""


class SystemUpdateStatus(IntEnum):
    """System Update Status"""
    OK = 0
    """Update OK"""
    FAILED_SCRIPT = 1
    """Script Failed"""
    FAILED_PREP = 10
    """Preparation Failed"""
    FAILED_FILENAME = 11
    """Filename Failed"""
    FAILED_DECRYPT = 12
    """Decryption Failed"""
    FAILED_UNPACK = 13
    """Unpacking Failed"""
    FAILED_VERIFY = 14
    """Verification Failed"""
    FAILED_MOVE = 15
    """Moving Failed"""


class TimeKeeperServiceStatus(IntEnum):
    """TimeKeeper Service Status"""
    STOPPED = 0
    """Service Stopped"""
    STARTED = 1
    """Service Started"""
    NA = 2
    """Not Available"""


class TimeKeeperServiceAction(IntEnum):
    """TimeKeeper Service Action"""
    STOP = 0
    """Stop"""
    START = 1
    """Start"""
    RESTART = 2
    """Restart"""


class CustomDefaultCommand(IntEnum):
    """Custom Default Command"""
    SET = 0
    """Set Custom Default"""
    CLEAR = 1
    """Clear Custom Default"""


class CustomDefaultScope(IntEnum):
    """Custom Default Scope"""
    ALL = 0
    """All"""
    INSTANCE = 1
    """Instance"""



class TrafficError(IntEnum):
    """Traffic Error"""
    NOT_PREPARED = 0
    """Not Prepared"""
    RATE_LENGTH_ERROR = 1
    """Rate Length Error"""
    PREPARED_OK = 2
    """Prepared OK"""


class TrafficEngine(IntEnum):
    """Traffic Engine"""
    TGA = 1
    """Normal TGA"""
    MICRO_TGA = 2
    """Micro TGA"""


#: L47 Enums
class Role(IntEnum):
    """L47 port role"""
    CLIENT = 0
    """Client"""
    SERVER = 1
    """Server"""


class Timescale(IntEnum):
    """Time scale"""
    MSECS = 0
    """Milliseconds"""
    SECONDS = 1
    """Seconds"""
    MINUTES = 2
    """Minutes"""
    HOURS = 3
    """Hours"""


class MSSType(IntEnum):
    """TCP Maximum Segment Type"""
    FIXED = 0
    """Fixed"""
    INCREMENT = 1
    """Incrementing"""
    RANDOM = 2
    """Pseudorandom"""


class RTOType(IntEnum):
    """TCP RTO Type"""
    STATIC = 0
    """Statis"""
    DYNAMIC = 1
    """Dynamic"""


class CongestionType(IntEnum):
    """TCP Congestion Control Algorithm"""
    NONE = 0
    """No TCP Congestion Control"""
    RENO = 1
    """RENO"""
    NEW_RENO = 2
    """New RENO"""


class IsEnabled(IntEnum):
    DISABLE = 0
    ENABLE = 1


class AlgorithmMethod(IntEnum):
    """Algorithm to calculate the TCP initial congestion window (ICWND)"""
    RFC5681 = 0
    """RFC 5681"""
    RFC2581 = 1
    """RFC 2581"""
    FIXED_FACTOR = 2
    """Fixed"""


class AutoOrManual(IntEnum):
    AUTOMATIC = 0
    MANUAL = 1


class EmbedIP(IntEnum):
    """Should Embed IP in MAC"""
    DONT_EMBED_IP = 0
    """Do not embed"""
    EMBED_IP = 1
    """Embed IP in MAC"""


class ApplicationLayerBehavior(IntEnum):
    """L47 Test Application Type"""
    NONE = 0
    """TCP connections are created according to the client and server ranges, and ramped up/down as specified in the load profile. But no payload is transmitted."""
    RAW = 1
    """Differs from ``NONE`` in that it transmits payload when the TCP connection is established."""
    REPLAY = 2
    """PCAP replay."""


class TrafficScenario(IntEnum):
    """Traffic direction scenario"""
    DOWNLOAD = 0
    """Server transmits payload to client."""
    UPLOAD = 1
    """Client transmits payload to server."""
    BOTH = 2
    """Payload is transmitted in both directions."""
    ECHO = 3
    """Client transmits payload to server, server echoes the payload back"""


class PayloadGenerationMethod(IntEnum):
    """Payload generation method."""
    FIXED = 0
    """Payload has a fixed value."""
    INCREMENT = 1
    """Payload consist of incrementing bytes."""
    RANDOM = 2
    """Payload consists of pseudo random bytes with a repeat cycle of 1 MB."""
    LONGRANDOM = 3
    """Payload consists of pseudo random bytes with a repeat cycle of 4 GB."""


class InfiniteOrFinite(IntEnum):
    INFINITE = 0
    FINITE = 1


class WhoClose(IntEnum):
    """How to close TCP connection when all payload has been transmitted."""
    NONE = 0
    """Keep the connection open after last byte is transmitted"""
    CLIENT = 1
    """Client closes the connection after last byte is receiver/transmitted"""
    SERVER = 2
    """Server closes the connection after last byte is transmitted"""


class LifecycleMode(IntEnum):
    """Connection lifecyle mode"""
    ONCE = 0
    """Connections are established during the ramp-up phase and not closed until the ramp-down phase of the load profile. That is, each configured connection only exists once."""
    IMMORTAL = 1
    """Connections are established during the ramp-up phase of the load profile, and are closed after the configured lifetime (configured by  P4G_RAW_CONN_LIFETIME). As connections close, new connections are established, attempting to keep the concurrent number of established connections constant. A new connection will have the same IP address as the connection it replaces, but will have a new TCP port number. This will simulate that the user (defined by the client IP address) is living on, and creates new connections as old connections close."""
    REINCARNATE = 2
    """Connections are established during the ramp-up phase of the load profile, and are closed after the configured lifetime (configured by  P4G_RAW_CONN_LIFETIME). As connections close, new connections are established, attempting to keep the concurrent number of established connections constant. A new connection will have the same TCP port number as the connection it replaces, but will have a new IP address. This will simulate that the user (defined by the client IP address) ceases to exist, and new users appear as old users die."""


class L47IPVersion(IntEnum):
    """IP version of the Connection Group"""
    IPV4 = 4
    """IPv4"""
    IPV6 = 6
    """IPv6"""


class L47ProtocolType(IntEnum):
    """L4 protocol of the Connection Group"""
    TCP = 0
    """TCP"""
    UDP = 1
    """UDP"""


class L47TrafficState(IntEnum):
    """L47 traffic state"""
    OFF = 0
    """Off"""
    ON = 1
    """On"""
    STOP = 2
    """Stop"""
    PREPARE = 3
    """Prepare"""
    PRERUN = 4
    """Prerun"""


class L47PortState(IntEnum):
    """L47 port state"""
    OFF = 0
    """Off"""
    PREPARE = 1
    """Prepare"""
    PREPARE_RDY = 2
    """Prepare Ready"""
    PREPARE_FAIL = 3
    """Prepare Failed"""
    PRERUN = 4
    """Prerun"""
    PRERUN_RDY = 5
    """Prepare Ready"""
    RUNNING = 6
    """Running"""
    STOPPING = 7
    """Stopping"""
    STOPPED = 8
    """Stopped"""


class L47PortSpeed(IntEnum):
    """L47 port speed mode"""
    AUTO = 0
    """Auto"""
    F100M = 1
    """100 Mbit/s"""
    F1G = 2
    """1 Gbit/s"""
    F2_5G = 3
    """2.5 Gbit/s"""
    F5G = 4
    """5 Gbit/s"""
    F10G = 5
    """10 Gbit/s"""
    F25G = 6
    """25 Gbit/s"""
    F40G = 7
    """40 Gbit/s"""
    F50G = 8
    """50 Gbit/s"""
    F100G = 9
    """100 Gbit/s"""


class CaptureSize(IntEnum):
    """Capture size"""
    FULL = 0
    """Capture whole packets"""
    SMALL = 1
    """Capture truncated packets"""


class ReplayParserState(IntEnum):
    OFF = 0
    PARSING = 1


class IsPresent(IntEnum):
    NOT_PRESENT = 0
    PRESENT = 1


class LicenseSpeed(IntEnum):
    UNDEFINED = 0
    F1G = 1
    F2_5G = 2
    F5G = 3
    F10G = 4
    F25G = 5
    F40G = 6
    F50G = 7
    F100G = 8


class TLSVersion(IntEnum):
    """TLS protocol version"""
    SSLV3 = 0
    """SSL v3"""
    TLS10 = 1
    """TLS v1.0"""
    TLS11 = 2
    """TLS v1.1"""
    TLS12 = 3
    """TLS v1.2"""


class ResourceAllocationMode(IntEnum):
    SIMPLE = 0
    ADVANCED = 1


class ReplaySchedulingMode(IntEnum):
    BANDWIDTH = 0
    TIME = 1


class ReplaySyncBasedOn(IntEnum):
    PER_CONN = 0
    PER_USER = 1


#: Impairment Enums
class CorruptionType(IntEnum):
    """Impairment corruption type"""
    OFF = 0
    """Off"""
    ETH = 1
    """Ethernet"""
    IP = 2
    """IP"""
    UDP = 3
    """UDP"""
    TCP = 4
    """TCP"""
    BER = 5
    """Bit Error Rate"""


class PolicerMode(IntEnum):
    """Policer mode"""
    L1 = 0
    """Policer performed at Layer 1 level. I.e. including the preamble and min interpacket gap."""
    L2 = 1
    """Policer performed at Layer 2 level. I.e. excluding the preamble and min interpacket gap"""


class FilterUse(IntEnum):
    """Use of filter."""
    OFF = 0
    """No filtering will be done"""
    AND = 1
    """Filtering will be done"""


class InfoAction(IntEnum):
    """Action of filter."""
    EXCLUDE = 0
    """Matching packets are excluded from the flow"""
    INCLUDE = 1
    """Matching packets are included from the flow"""


class L2PlusPresent(IntEnum):
    """Presence of Layer-2+ protocols"""
    NA = 0
    """No Layer 2+ protocols"""
    VLAN1 = 1
    """One VLAN Tag is present"""
    VLAN2 = 2
    """Two VLAN Tags are present"""
    MPLS = 3
    """MPLS label is present"""


class L3PlusPresent(IntEnum):
    """Presence of Layer-3 protocols"""
    NA = 0
    """No Layer 3 protocols"""
    IP4 = 1
    """IPv4 is present"""
    IP6 = 2
    """IPv6 is present"""


class FilterMode(IntEnum):
    """Impairment Filter Mode"""
    BASIC = 0
    """Basic Mode"""
    EXTENDED = 1
    """Extended Mode"""


class ImpairmentLatencyMode(IntEnum):
    """Impairment Latency Mode"""
    NORMAL = 0
    """Normal"""
    EXTENDED = 1
    """Extended"""


class ShadowWorkingSelection(IntEnum):
    """Shadow Working Selection"""
    SHADOW = 0
    """Shadow"""
    WORKING = 1
    """Working"""

class FilterType(IntEnum):
    """Filter Type for Impairemennt"""
    SHADOW = 0 
    """Shadow Copy"""
    WORKING = 1
    """Wokring Copy"""


class FilterVlanType(IntEnum):
    """VLAN PCP Settings for VLAN Filter"""
    INNER = 0 
    """VLAN1 (0) (INNER VLAN Tag is specified for the filter – used also when only 1 VLAN), indicates single/inner VLAN-TPID=0x8100"""
    OUTER = 1 
    """VLAN2 (1) (OUTER VLAN Tag is specified for the filter), indicates outer VLAN-TPID=0x88A8"""


class LatencyTypeCustomDist(IntEnum):
    """Latency Type for Custom Distribution"""
    INTERPACKET_DISTRIBUTION = 0
    """Interpacket Distribution"""
    LATENCY_DISTRIBUTION = 1
    """Latency Distribution"""


class ImpairmentTypeIndex(IntEnum):
    """Impairment Type Index"""
    DROP = 0
    """Drop"""
    MISORDER = 1
    """Misorder"""
    DELAYJITTER = 2
    """Delay/Jitter"""
    DUPLICATION = 3
    """Duplication"""
    CORRUPTION = 4
    """Corruption"""
    POLICER = 5
    """Policer"""
    SHAPER = 6
    """Shaper"""


#: TSN Enums
class TSNConfigProfile(IntEnum):
    """TSN PTP Configuration profile"""
    AUTOMOTIVE = 0
    """Defaults suitable for automotive testing"""
    IEEE1588V2 = 1
    """Defaults suitable for PTP  testing"""


class TSNPortRole(IntEnum):
    """TSN port role"""
    GRANDMASTER = 0
    """Grandmaster role"""
    SLAVE = 1
    """Slave role"""


class TSNDeviationMode(IntEnum):
    FIXED = 0


class TSNTimeSource(IntEnum):
    """TSN time source"""
    ATOMIC = 0x10
    """Atomic"""
    GPS = 0x20
    """GPS"""
    TERRESTRIAL = 0x30
    """Terrestrial"""
    PTP = 0x40
    """PTP"""
    NTP = 0x50
    """NTP"""
    HAND_SET = 0x60
    """Handset"""
    OTHER = 0x90
    """Other"""
    INTERNAL_OSC = 0xA0
    """Internal oscillator"""


class TSNHistogramSource(IntEnum):
    """Data source for TSN histogram"""
    DRIFT = 0
    """Post-servo offset to Grandmaster"""
    DRIFTPRE = 1
    """Pre-servo offset to Grandmaster"""
    PDELAY = 2
    """PDelay data."""
    NRR = 3
    """Neighbor Rate Ratio data."""


class TSNStatisticsTypes(IntEnum):
    """TSN Statistics Types"""
    ALL = 0
    """All."""
    PACKETCOUNT = 1
    """Packet count."""
    OFFSET = 2
    """Offset."""
    PDELAY = 3
    """PDelay."""
    SYNCRATE = 4
    """Sync rate."""


class PPMSweepMode(IntEnum):
    """Module clock PPM Sweep Modes"""
    OFF = 0
    """Off"""
    TRIANGLE = 1
    """Triangle sweeping"""

class PPMSweepStatus(IntEnum):
    """Module clock PPM Sweep Status"""
    OFF = 0
    """Off"""
    SWEEPING = 1
    """The module is sweeping"""

class ReconciliationSublayerSupport (IntEnum):
    """Reconciliation Sublayer Support"""
    NO_SUPPORT = 0
    """Not Supported"""
    FAULT_SIGNALING = 1
    """Supported, which means P_FAULTSTATUS and P_FAULTSIGNALLING are supported by the port."""


import functools
from typing import TYPE_CHECKING
from .bases.port_l23 import BasePortL23
from .bases.port_reception_statistics import PortReceptionStatistics
from .bases.port_transmission_statistics import PortTransmissionStatistics

from xoa_driver.internals.core.commands import (
    P_MDIXMODE,
    # P_ENGINENAMES, #TODO: need to implement
    # P_ENGINELOAD, #TODO: need to implement
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as idx_mgr
from xoa_driver.internals.indices.streams.base_stream import BaseStreamIdx
from xoa_driver.internals.indices.filter.base_filter import BaseFilterIdx
from xoa_driver.internals.state_storage import ports_state

VEStreamIndices = idx_mgr.IndexManager[BaseStreamIdx]
VEFilterIndices = idx_mgr.IndexManager[BaseFilterIdx]

class Engine:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        # self.names = P_ENGINENAMES(conn, module_id, port_id)
        # self.load = P_ENGINELOAD(conn, module_id, port_id)
        ...

class PortStatistics:
    """L23 VE port statistics"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = PortReceptionStatistics(conn, module_id, port_id)
        """L23 VE port's RX statistics."""
        
        self.rx = PortTransmissionStatistics(conn, module_id, port_id)
        """L23 VE port's TX statistics."""

class PortL23VE(BasePortL23):
    """L23 VE port"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)

        self._local_states = ports_state.PortL23LocalState()
        
        self.mdix_mode = P_MDIXMODE(conn, module_id, port_id)
        """MDI/MDIX mode.
        Representation of P_MDIXMODE
        """

        self.engine = Engine(conn, module_id, port_id)
        """Engine is not supported yet."""
        
        self.statistics = PortStatistics(conn, module_id, port_id)
        
        self.streams: VEStreamIndices = idx_mgr.IndexManager(
            conn, 
            BaseStreamIdx, 
            module_id, 
            port_id
        )
        """L23 VE port's stream index manager."""

        self.filters: VEFilterIndices = idx_mgr.IndexManager(
            conn, 
            BaseFilterIdx, 
            module_id, 
            port_id
        )
        """L23 VE port's filter index manager."""
    
    @property
    def info(self) -> ports_state.PortL23LocalState:
        return self._local_states
    
    async def _setup(self):
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        return self
    
    on_mdix_mode_change = functools.partialmethod(utils.on_event, P_MDIXMODE)
    """Register a callback to the event that the port's MDI/MDIX mode changes."""


import functools
import typing
from xoa_driver.internals.ports import base_port
from xoa_driver.internals.core.commands import (
    P_CAPABILITIES,
    P_SPEED,
    P_SPEEDREDUCTION,
    P_INTERFRAMEGAP,
    P_MACADDRESS,
    P_IPADDRESS,
    P_ARPREPLY,
    P_PINGREPLY,
    P_PAUSE,
    P_RANDOMSEED,
    P_LOOPBACK,
    P_TRAFFIC,
    P_CAPTURE,
    P_XMITONE,
    P_LATENCYOFFSET,
    P_LATENCYMODE,
    P_AUTOTRAIN,
    P_MIXWEIGHTS,
    P_TRAFFICERR,
    P_GAPMONITOR,
    P_CHECKSUM,
    P_MIXLENGTH,
    P_ARPRXTABLE,
    P_NDPRXTABLE,
    P_MULTICAST,
    P_MULTICASTEXT,
    P_MCSRCLIST,
    P_MULTICASTHDR,
    P_TXMODE,
    P_RATEFRACTION,
    P_RATEPPS,
    P_RATEL2BPS,
    P_PAYLOADMODE,
    P_TXENABLE,
    P_MAXHEADERLENGTH,
    P_TXTIMELIMIT,
    P_TXTIME,
    P_XMITONETIME,
    P_IPV6ADDRESS,
    P_ARPV6REPLY,
    P_PINGV6REPLY,
    P_ERRORS,
    P_TXPREPARE,
    P_TXDELAY,
    P_TPLDMODE,
    P_TXPACKETLIMIT,
    P_PFCENABLE,
    P_TXBURSTPERIOD,
)
if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.core.transporter import funcs
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as idx_mgr
from xoa_driver.internals.state_storage import ports_state
from xoa_driver.internals.indices.length_term import LengthTermIdx
from xoa_driver.internals.indices.match_term import MatchTermIdx

from .port_capture import PortCapture

LengthTermIndices = idx_mgr.IndexManager[LengthTermIdx]
MatchTermIndices = idx_mgr.IndexManager[MatchTermIdx]

class TxSinglePacket:
    """L23 port single-packet transmission"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.send = P_XMITONE(conn, module_id, port_id)
        """Send one packet from the L23 port without a stream config.
        Representation of P_XMITONE
        """
        
        self.time = P_XMITONETIME(conn, module_id, port_id)
        """The time at which the latest packet was transmitted using the P_XMITONE` command.
        Representation of P_XMITONETIME
        """


class TxConfiguration:
    """L23 port TX configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mode = P_TXMODE(conn, module_id, port_id)
        """L23 port TX mode.
        Representation of P_TXMODE
        """
        
        self.enable = P_TXENABLE(conn, module_id, port_id)
        """Enabling L23 port TX.
        Representation of P_TXENABLE
        """
        
        self.time_limit = P_TXTIMELIMIT(conn, module_id, port_id)
        """L23 port TX time limit.
        Representation of P_TXTIMELIMIT
        """
        
        self.time = P_TXTIME(conn, module_id, port_id)
        """L23 port TX time.
        Representation of P_TXTIME
        """
        
        self.repare = P_TXPREPARE(conn, module_id, port_id)
        """Prepare L23 port for transmission.
        Representation of P_TXPREPARE
        """
        
        self.delay = P_TXDELAY(conn, module_id, port_id)
        """L23 port TX delay.
        Representation of P_TXDELAY
        """
        
        self.packet_limit = P_TXPACKETLIMIT(conn, module_id, port_id)
        """L23 port TX packet limit
        Representation of P_TXPACKETLIMIT
        """
        
        self.burst_period = P_TXBURSTPERIOD(conn, module_id, port_id)
        """L23 port TX burst period.
        Representation of P_TXBURSTPERIOD
        """


class Rate:
    """L23 port TX rate"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.fraction = P_RATEFRACTION(conn, module_id, port_id)
        """L23 port rate in ppm.
        Representation of P_RATEFRACTION
        """
        
        self.pps = P_RATEPPS(conn, module_id, port_id)
        """L23 port rate in packets per second.
        Representation of P_RATEPPS
        """
        
        self.l2_bps = P_RATEL2BPS(conn, module_id, port_id)
        """L23 port rate in L2 bits per second.
        Representation of P_RATEL2BPS
        """


class Multicast:
    """L23 port multicast configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mode = P_MULTICAST(conn, module_id, port_id)
        """L23 port multicast mode.
        Representation of P_MULTICAST
        """
        
        self.mode_extended = P_MULTICASTEXT(conn, module_id, port_id)
        """L23 port multicast extended mode.
        Representation of P_MULTICASTEXT
        """
        
        self.source_list = P_MCSRCLIST(conn, module_id, port_id)
        """L23 port multicast source list.
        Representation of P_MCSRCLIST
        """
        
        self.header = P_MULTICASTHDR(conn, module_id, port_id)
        """L23 port multicast IGMP header.
        Representation of P_MULTICASTHDR
        """


class IPv4:
    """L23 port IPv4 configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.address = P_IPADDRESS(conn, module_id, port_id)
        """L23 port IPv4 address.
        Representation of P_IPADDRESS
        """
        
        self.arp_reply = P_ARPREPLY(conn, module_id, port_id)
        """L23 port reply to ARP request.
        Representation of P_ARPREPLY
        """
        
        self.ping_reply = P_PINGREPLY(conn, module_id, port_id)
        """L23 port reply to PING request.
        Representation of P_PINGREPLY
        """


class IPv6:
    """L23 port IPv6 configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.address = P_IPV6ADDRESS(conn, module_id, port_id)
        """L23 port IPv6 address.
        Representation of P_IPV6ADDRESS
        """
        
        self.arp_reply = P_ARPV6REPLY(conn, module_id, port_id)
        """L23 port reply to NDP Neighbor Solicitation.
        Representation of P_ARPV6REPLY
        """
        
        self.ping_reply = P_PINGV6REPLY(conn, module_id, port_id)
        """L23 port reply to PINGv6 request.
        Representation of P_PINGV6REPLY
        """


class NetworkConfiguration: # will be extended in genuine ports
    """L23 port networking configuration"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mac_address = P_MACADDRESS(conn, module_id, port_id)
        """L23 port MAC address.
        Representation of P_MACADDRESS
        """
        
        self.ipv4 = IPv4(conn, module_id, port_id)
        """L23 port IPv4 address configuration.
        """
        
        self.ipv6 = IPv6(conn, module_id, port_id)
        """L23 port IPv6 address configuration.
        """


class LatencyConfiguration:
    """L23 port latency configuration"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.offset = P_LATENCYOFFSET(conn, module_id, port_id)
        """L23 port latency offset.
        Representation of P_LATENCYOFFSET
        """
        
        self.mode = P_LATENCYMODE(conn, module_id, port_id)
        """L23 port latency measurement mode.
        Representation of P_LATENCYMODE
        """

class Mix:
    """L23 port IMIX configuration"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.weights = P_MIXWEIGHTS(conn, module_id, port_id)
        """L23 port IMIX weights
        Representation of P_MIXWEIGHTS
        """
        
        self.lengths = tuple(
            P_MIXLENGTH(conn, module_id, port_id, idx)
            for idx in range(16)
        ) #TODO: need to add manager for handle specific indices only
        """L23 port IMIX lengths.
        Representation of P_MIXLENGTH
        """

class Speed:
    """L23 port speed configuration"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.current = P_SPEED(conn, module_id, port_id)
        """L23 port current speed in units of Mbps.
        Representation of P_SPEED
        """
        
        self.reduction = P_SPEEDREDUCTION(conn, module_id, port_id)
        """L23 port speed reduction in ppm.
        Representation of P_SPEEDREDUCTION
        """

class Traffic:
    """L23 port traffic generation"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.state = P_TRAFFIC(conn, module_id, port_id)
        """L23 port traffic status and action.
        Representation of P_TRAFFIC
        """
        
        self.error = P_TRAFFICERR(conn, module_id, port_id)
        """L23 port traffic error.
        Representation of P_TRAFFICERR
        """

class BasePortL23(base_port.BasePort[ports_state.PortL23LocalState]):
    """L23 port layout which is relevant to all L23 ports."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.capabilities = P_CAPABILITIES(conn, module_id, port_id)
        """L23 port capabilities.
        Representation of P_CAPABILITIES
        """
        
        self.pause = P_PAUSE(conn, module_id, port_id)
        """L23 port response to Ethernet PAUSE frames.
        Representation of P_PAUSE
        """
        
        self.loop_back = P_LOOPBACK(conn, module_id, port_id)
        """L23 port loopback mode.
        Representation of P_LOOPBACK
        """
        
        
        self.errors_count = P_ERRORS(conn, module_id, port_id)
        """L23 port errors.
        Representation of P_ERRORS
        """

        self.interframe_gap = P_INTERFRAMEGAP(conn, module_id, port_id)
        """L23 port interframe gap.
        Representation of P_INTERFRAMEGAP
        """
        
        self.max_header_length = P_MAXHEADERLENGTH(conn, module_id, port_id)
        """L23 port maximum header length.
        Representation of P_MAXHEADERLENGTH
        """
        
        self.tpld_mode = P_TPLDMODE(conn, module_id, port_id)
        """L23 port test payload mode.
        Representation of P_TPLDMODE
        """
        
        self.pfc_enable = P_PFCENABLE(conn, module_id, port_id)
        """L23 port Ethernet Priority Flow Control (PFC).
        Representation of P_PFCENABLE
        """
        
        self.random_seed = P_RANDOMSEED(conn, module_id, port_id)
        """L23 port seed value.
        Representation of P_RANDOMSEED
        """
        
        self.payload_mode = P_PAYLOADMODE(conn, module_id, port_id)
        """L23 port payload mode.
        Representation of P_PAYLOADMODE
        """
        
        self.autotrain = P_AUTOTRAIN(conn, module_id, port_id)
        """L23 port interval between auto training packets.
        Representation of P_AUTOTRAIN
        """
        
        self.gap_monitor = P_GAPMONITOR(conn, module_id, port_id)
        """L23 port gap monitor.
        Representation of P_GAPMONITOR
        """
        
        self.checksum = P_CHECKSUM(conn, module_id, port_id)
        """L23 port extra payload integrity checksum.
        Representation of P_CHECKSUM
        """

        self.arp_rx_table = P_ARPRXTABLE(conn, module_id, port_id)
        """L23 port ARP table.
        Representation of P_ARPRXTABLE
        """
        
        self.ndp_rx_table = P_NDPRXTABLE(conn, module_id, port_id)
        """L23 port NDP table.
        Representation of P_NDPRXTABLE
        """

        self.capturer = PortCapture(conn, module_id, port_id)
        """L23 port capturer configuration."""
        
        self.speed = Speed(conn, module_id, port_id)
        """L23 port speed configuration."""
        
        self.traffic = Traffic(conn, module_id, port_id)
        """L23 port traffic configuration."""
        
        self.mix = Mix(conn, module_id, port_id)
        """L23 port IMIX configuration."""
        
        self.latency_config = LatencyConfiguration(conn, module_id, port_id)
        """L23 port latency configuration."""
        
        self.rate = Rate(conn, module_id, port_id)
        """L23 port rate."""
        
        self.tx_config = TxConfiguration(conn, module_id, port_id)
        """L23 port TX configuration."""
        
        self.tx_single_pkt = TxSinglePacket(conn, module_id, port_id)
        """L23 port single-packet TX configuration."""
        
        self.multicast = Multicast(conn, module_id, port_id)
        """L23 port multicast configuration."""
        
        self.net_config = NetworkConfiguration(conn, module_id, port_id)
        """L23 port network configuration."""

        self.local_states = ports_state.PortL23LocalState()
        """L23 port local states."""

        self.length_terms: LengthTermIndices = idx_mgr.IndexManager(
            conn,
            LengthTermIdx,
            module_id,
            port_id
        )
        """L23 port's length term index manager."""

        self.match_terms: MatchTermIndices = idx_mgr.IndexManager(
            conn,
            MatchTermIdx,
            module_id,
            port_id
        )
        """L23 port's match term index manager."""

    on_speed_change = functools.partialmethod(utils.on_event, P_SPEED)
    """Register a callback to the event that the port's speed changes."""

    on_speed_reduction_change = functools.partialmethod(utils.on_event, P_SPEEDREDUCTION)
    """Register a callback to the event that the port's speed reduction changes."""

    on_traffic_change = functools.partialmethod(utils.on_event, P_TRAFFIC)
    """Register a callback to the event that the port's traffic status changes."""

    on_capturer_state_change = functools.partialmethod(utils.on_event, P_CAPTURE)
    """Register a callback to the event that the port's capturer state changes."""

    on_tx_enable_change = functools.partialmethod(utils.on_event, P_TXENABLE)
    """Register a callback to the event that the port's TX status changes."""

    on_interframe_gap_change = functools.partialmethod(utils.on_event, P_INTERFRAMEGAP)
    """Register a callback to the event that the port's interframe gap configuration changes."""

    on_mac_address_change = functools.partialmethod(utils.on_event, P_MACADDRESS)
    """Register a callback to the event that the port's MAC address changes."""

    on_ipv4_address_change = functools.partialmethod(utils.on_event, P_IPADDRESS)
    """Register a callback to the event that the port's IPv4 address changes."""

    on_ipv4_arp_reply_change = functools.partialmethod(utils.on_event, P_ARPREPLY)
    """Register a callback to the event that the port's ARP reply setting changes."""

    on_ipv4_ping_reply_change = functools.partialmethod(utils.on_event, P_PINGREPLY)
    """Register a callback to the event that the port's PING reply setting changes."""

    on_multicast_mode_change = functools.partialmethod(utils.on_event, P_MULTICAST)
    """Register a callback to the event that the port's multicast configuration changes."""

    on_multicast_mode_extended_change = functools.partialmethod(utils.on_event, P_MULTICASTEXT)
    """Register a callback to the event that the port's extended multicast configuration changes."""

    on_multicast_header_change = functools.partialmethod(utils.on_event, P_MULTICASTHDR)
    """Register a callback to the event that the port's multicast packet header changes."""

    on_multicast_source_list_change = functools.partialmethod(utils.on_event, P_MCSRCLIST)
    """Register a callback to the event that the port's multicast source list changes."""

    on_ipv6_address_change = functools.partialmethod(utils.on_event, P_IPV6ADDRESS)
    """Register a callback to the event that the port's IPv6 address setting changes."""

    on_ipv6_arp_reply_change = functools.partialmethod(utils.on_event, P_ARPV6REPLY)
    """Register a callback to the event that the port's NDP reply setting changes."""

    on_ipv6_ping_reply_change = functools.partialmethod(utils.on_event, P_PINGV6REPLY)
    """Register a callback to the event that the port's PINGv6 reply setting changes."""

    on_arp_rx_table_change = functools.partialmethod(utils.on_event, P_ARPRXTABLE)
    """Register a callback to the event that the port's ARP table changes."""

    on_ndp_rx_table_change = functools.partialmethod(utils.on_event, P_NDPRXTABLE)
    """Register a callback to the event that the port's NDP table changes."""

    on_pause_change = functools.partialmethod(utils.on_event, P_PAUSE)
    """Register a callback to the event that the port's response to pause frame setting changes."""

    on_pfc_enable_change = functools.partialmethod(utils.on_event, P_PFCENABLE)
    """Register a callback to the event that the port's response to PFC setting changes."""

    on_random_seed_change = functools.partialmethod(utils.on_event, P_RANDOMSEED)
    """Register a callback to the event that the port's random seed changes."""

    on_latency_offset_change = functools.partialmethod(utils.on_event, P_LATENCYOFFSET)
    """Register a callback to the event that the port's latency offset changes."""

    on_latency_mode_change = functools.partialmethod(utils.on_event, P_LATENCYMODE)
    """Register a callback to the event that the port's latency measurement mode changes."""

    on_tx_time_limit_change = functools.partialmethod(utils.on_event, P_TXTIMELIMIT)
    """Register a callback to the event that the port's TX time limit changes."""

    on_tx_burst_period_change = functools.partialmethod(utils.on_event, P_TXBURSTPERIOD)
    """Register a callback to the event that the port's TX burst period changes."""

    on_tx_packet_limit_change = functools.partialmethod(utils.on_event, P_TXPACKETLIMIT)
    """Register a callback to the event that the port's TX packet count limit changes."""

    on_tx_mode_change = functools.partialmethod(utils.on_event, P_TXMODE)
    """Register a callback to the event that the port's TX mode changes."""

    on_tx_delay_change = functools.partialmethod(utils.on_event, P_TXDELAY)
    """Register a callback to the event that the port's TX deplay changes."""

    on_max_header_length_change = functools.partialmethod(utils.on_event, P_MAXHEADERLENGTH)
    """Register a callback to the event that the port's maximum packet header length changes."""

    on_auto_train_change = functools.partialmethod(utils.on_event, P_AUTOTRAIN)
    """Register a callback to the event that the port's auto training setting changes."""

    on_loop_back_change = functools.partialmethod(utils.on_event, P_LOOPBACK)
    """Register a callback to the event that the port's loopback mode changes."""

    on_checksum_change = functools.partialmethod(utils.on_event, P_CHECKSUM)
    """Register a callback to the event that the port's extra payload integrity checksum changes."""

    on_gap_monitor_change = functools.partialmethod(utils.on_event, P_GAPMONITOR)
    """Register a callback to the event that the port's gap-start and gap-stop settings change."""

    on_mix_weights_change = functools.partialmethod(utils.on_event, P_MIXWEIGHTS)
    """Register a callback to the event that the port's distribution of the MIX packet length changes."""

    on_mix_length_change = functools.partialmethod(utils.on_event, P_MIXLENGTH)
    """Register a callback to the event that the port's frame sizes defined for each mix weight changes."""

    on_tpld_mode_change = functools.partialmethod(utils.on_event, P_TPLDMODE)
    """Register a callback to the event that the port's TPLD mode changes."""

    on_payload_mode_change = functools.partialmethod(utils.on_event, P_PAYLOADMODE)
    """Register a callback to the event that the port's payload mode changes."""



import asyncio
import functools
from typing import TYPE_CHECKING
from xoa_driver.internals.ports import base_port

from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.core.commands import (
    P_CAPABILITIES,
    P_INTERFACE,
    P_STATUS,
    P_TXENABLE,
    P_LOADMODE,
    P_EMULATE,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    
from xoa_driver.internals.state_storage import ports_state
from ..pcs_pma_ijkl_chimera import PcsPma
from .pe_custom_distribution import CustomDistributions
from .port_emulation import ChimeraPE

class PortChimera(base_port.BasePort[ports_state.PortChimeraLocalState]):
    """Chimera port for traffic impairment."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.capabilities = P_CAPABILITIES(conn, module_id, port_id)
        """Chimera port capabilities.
        Representation of P_CAPABILITIES
        """

        self.interface = P_INTERFACE(conn, module_id, port_id)
        """Physical interface type of the Chimera port.
        Representation of P_INTERFACE
        """

        self.status = P_STATUS(conn, module_id, port_id)
        """Chimera port's received optical signal level'.
        Representation of P_STATUS
        """

        self.tx_enable = P_TXENABLE(conn, module_id, port_id)
        """Enabling Chimera port TX.
        Representation of P_TXENABLE
        """

        self.load_mode = P_LOADMODE(conn, module_id, port_id)
        """Load mode of the Chimera port.
        Representation of P_LOADMODE
        """

        self.emulation = ChimeraPE(self._conn, *self.kind) # TODO: Missing description 
        
        self.emulate = P_EMULATE(conn, module_id, port_id)
        """Chimera port emulation control.
        Representation of P_EMULATE
        """

        self.custom_distributions = CustomDistributions(conn, module_id, port_id)
        """Custom distributions."""
        
        self.pcs_pma = PcsPma(conn, self)
        """"PCS/PMA settings."""
        
        self._local_states = ports_state.PortChimeraLocalState()
        """Local states of the Chimera port."""
    
    @property
    def info(self) -> ports_state.PortChimeraLocalState:
        return self._local_states
    
    async def _setup(self):
        await asyncio.gather(
            self._local_states.initiate(self),
            self.custom_distributions.server_sync(),
            self.emulation
        )
        self._local_states.register_subscriptions(self)
        return self
    
    on_interface_change = functools.partialmethod(utils.on_event, P_INTERFACE)
    """Register a callback to the event that the port's interface type changes."""

    on_status_change = functools.partialmethod(utils.on_event, P_STATUS)
    """Register a callback to the event that the port's optical signal level changes."""

    on_emulate_change = functools.partialmethod(utils.on_event, P_EMULATE)
    """Register a callback to the event that the port's emulation state changes."""
    
    on_tx_enable_change = functools.partialmethod(utils.on_event, P_TXENABLE)
    """Register a callback to the event that the port's TX status changes."""


import functools
from typing import TYPE_CHECKING
from .bases.port_l23_genuine import BasePortL23Genuine

from xoa_driver.internals.core.commands import (
    P_FAULTSIGNALING,
    P_FAULTSTATUS,
    P_DYNAMIC,
)
from xoa_driver.internals.utils import attributes as utils
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from .pcs_pma_ghijkl import (
    PcsPma,
    SerDes,
)
class Fault:
    """L23 port fault settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.signaling = P_FAULTSIGNALING(conn, module_id, port_id)
        """L23 port fault signaling.
        Representation of P_FAULTSIGNALING
        """
        
        self.status = P_FAULTSTATUS(conn, module_id, port_id)
        """L23 port fault status.
        Representation of P_FAULTSTATUS
        """


class FamilyG(BasePortL23Genuine):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.dynamic = P_DYNAMIC(conn, module_id, port_id)
        """L23 port's dynamic traffic change.
        Representation of P_DYNAMIC
        """
        
        self.pcs_pma = PcsPma(conn, self)
        """PCS/PMA settings."""

    async def _setup(self):
        await super()._setup()
        self.ser_des = tuple(
            SerDes(self._conn, *self.kind, serdes_xindex=serdes_xindex)
            for serdes_xindex in range(self.info.capabilities.serdes_count)
        )
        return self

    on_fault_signaling_change = functools.partialmethod(utils.on_event, P_FAULTSIGNALING)
    """Register a callback to the event that the port's fault signalling changes."""
    
    on_dynamic_change = functools.partialmethod(utils.on_event, P_DYNAMIC)
    """Register a callback to the event that the port's dynamic traffic setting changes."""



class PLoki100G3S1P(FamilyG):
    """L23 port on Loki-100G-3S-1P module.
    """
    ...

class PLoki100G3S1P_b(FamilyG):
    """L23 port on Loki-100G-3S-1P[b] module.
    """
    ...

class PLoki100G3S1PSE(FamilyG):
    """L23 port on Loki-100G-3S-1P-SE module.
    """
    ...




import functools
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core.commands import (
    P_FLASH,
    P_SPEEDSELECTION,
    P_SPEEDS_SUPPORTED,
    P_UAT_MODE,
    P_UAT_FLR,
    P_STATUS,
    P_LPSUPPORT,
    P_TCVRSTATUS,
    P_TXRUNTLENGTH,
    P_RXRUNTLENGTH,
    P_RXRUNTLEN_ERRS,
    P_TXPREAMBLE_REMOVE,
    P_RXPREAMBLE_INSERT,
    P_LOADMODE,
    PP_FECMODE,
)
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as idx_mgr
from xoa_driver.internals.indices.streams.genuine_stream import GenuineStreamIdx
from xoa_driver.internals.indices.filter.genuine_filter import GenuineFilterIdx
from xoa_driver.internals.indices.port_dataset import PortDatasetIdx
from xoa_driver.internals.state_storage import ports_state

from .port_l23 import (
    BasePortL23,
    Speed,
    TxConfiguration,
)

from .port_transceiver import PortTransceiver
from .port_reception_statistics import GenuinePortReceptionStatistics
from .port_transmission_statistics import PortTransmissionStatistics

StreamIndices = idx_mgr.IndexManager[GenuineStreamIdx]
FilterIndices = idx_mgr.IndexManager[GenuineFilterIdx]
PortDatasetIndices = idx_mgr.IndexManager[PortDatasetIdx]


class SpeedMode:
    """L23 port's speed mode"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.selection = P_SPEEDSELECTION(conn, module_id, port_id)
        """L23 port speed mode selection.
        Representation of P_SPEEDSELECTION
        """
        
        self.supported = P_SPEEDS_SUPPORTED(conn, module_id, port_id)
        """L23 port's supported speed modes.
        Representation of P_SPEEDS_SUPPORTED
        """

class GenuineSpeed(Speed):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.mode = SpeedMode(conn, module_id, port_id)
        """L23 port's speed mode."""

class UnAvailableTime:
    """UnAvailable Time"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mode = P_UAT_MODE(conn, module_id, port_id)
        """L23 port's Unvailable Time mode.
        Representation of P_UAT_MODE
        """
        
        
        self.frame_loss_ratio = P_UAT_FLR(conn, module_id, port_id)
        """L23 port's Frame Loss Ratio for UAT.
        Representation of P_UAT_FLR
        """


class GenuineTxConfiguration(TxConfiguration):
    """L23 port's TX configuration."""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.runt_length = P_TXRUNTLENGTH(conn, module_id, port_id)
        """L23 port's TX runt length.
        Representation of P_TXRUNTLENGTH
        """
        
        self.preamble_remove = P_TXPREAMBLE_REMOVE(conn, module_id, port_id)
        """L23 port's removal of preamble from outgoing packets.
        Representation of P_TXPREAMBLE_REMOVE
        """


class GenuineRx:
    """L23 port's RX configuration."""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.runt_length = P_RXRUNTLENGTH(conn, module_id, port_id)
        """L23 port's RX runt length.
        Representation of P_RXRUNTLENGTH
        """
        
        self.preamble_insert = P_RXPREAMBLE_INSERT(conn, module_id, port_id)
        """L23 port's insertion of preamble into incoming packets.
        Representation of P_RXPREAMBLE_INSERT
        """
        
        self.runt_length_errors = P_RXRUNTLEN_ERRS(conn, module_id, port_id)
        """L23 port's RX runt length errors..
        Representation of P_RXRUNTLEN_ERRS
        """


class PortStatistics:
    """L23 port statistics"""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = GenuinePortReceptionStatistics(conn, module_id, port_id)
        """L23 port's RX statistics."""
        
        self.tx = PortTransmissionStatistics(conn, module_id, port_id)
        """L23 port's TX statistics."""


class BasePortL23Genuine(BasePortL23):
    """L23 port basic configuration."""
    
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        
        self._local_states = ports_state.PortL23GenuineLocalState()
        
        self.flash = P_FLASH(conn, module_id, port_id)
        """L23 port flashes.
        Representation of P_FLASH
        """
        
        self.status = P_STATUS(conn, module_id, port_id)
        """L23 port's received optical signal level'.
        Representation of P_STATUS
        """
        
        self.config_load_mode = P_LOADMODE(conn, module_id, port_id)
        """Chimera Port loading mode.
        Representation of P_LOADMODE
        """
        
        self.lp_support = P_LPSUPPORT(conn, module_id, port_id)
        """L23 port EEE capabilities.
        Representation of P_LPSUPPORT
        """
        
        self.tcvr_status = P_TCVRSTATUS(conn, module_id, port_id)
        """L23 port transceiver status information.
        Representation of P_TCVRSTATUS
        """
        
        self.fec_mode = PP_FECMODE(conn, module_id, port_id)
        """L23 port FEC mode.
        Representation of PP_FECMODE
        """

        self.speed = GenuineSpeed(conn, module_id, port_id)
        """L23 port speed configuration.
        """
        
        self.uat = UnAvailableTime(conn, module_id, port_id)
        """L23 port UnAvailable Time configuration.
        """
        
        self.tx_config = GenuineTxConfiguration(conn, module_id, port_id)
        """L23 port TX configuration.
        """

        self.transceiver = PortTransceiver(conn, module_id, port_id)
        """L23 port transceiver configuration.
        """
    
        self.statistics = PortStatistics(conn, module_id, port_id)
        """L23 port statistics.
        """

        self.streams: StreamIndices = idx_mgr.IndexManager(
            conn,
            GenuineStreamIdx,
            module_id,
            port_id
        )
        """L23 port stream index manager.
        """
        
        self.filters: FilterIndices = idx_mgr.IndexManager(
            conn,
            GenuineFilterIdx,
            module_id,
            port_id
        )
        """L23 port filter index manager.
        """
        
        self.datasets: PortDatasetIndices = idx_mgr.IndexManager(
            conn,
            PortDatasetIdx,
            module_id,
            port_id
        )
        """L23 port histogram index manager.
        """

    @property
    def info(self) -> ports_state.PortL23GenuineLocalState:
        return self._local_states
    
    async def _setup(self):
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        return self
    
    on_speed_selection_change = functools.partialmethod(utils.on_event, P_SPEEDSELECTION)
    """Register a callback to the event that the port's speed mode changes."""

    on_status_change = functools.partialmethod(utils.on_event, P_STATUS)
    """Register a callback to the event that the port's RX signal level (for optical ports) changes."""

    on_lp_support_change = functools.partialmethod(utils.on_event, P_LPSUPPORT)
    """Register a callback to the event that the port's Energy Efficient Ethernet (EEE) status changes."""

    on_uat_mode_change = functools.partialmethod(utils.on_event, P_UAT_MODE)
    """Register a callback to the event that the port's UnAvailable Time mode changes."""

    on_uat_flr_change = functools.partialmethod(utils.on_event, P_UAT_FLR)
    """Register a callback to the event that the port's threshold for the Frame Loss Ratio changes. Used in connection with UnAvailable Time."""

    on_flash_change = functools.partialmethod(utils.on_event, P_FLASH)
    """Register a callback to the event that the port's flash status changes."""
    
    on_tx_runt_length_change = functools.partialmethod(utils.on_event, P_TXRUNTLENGTH)
    """Register a callback to the event that the port's TX runt length setting changes."""

    on_rx_runt_length_change = functools.partialmethod(utils.on_event, P_RXRUNTLENGTH)
    """Register a callback to the event that the port's RX runt length detection changes."""

    on_tx_preamble_remove_change = functools.partialmethod(utils.on_event, P_TXPREAMBLE_REMOVE)
    """Register a callback to the event that the port's preamble removal status changes."""

    on_rx_preamble_insert_change = functools.partialmethod(utils.on_event, P_RXPREAMBLE_INSERT)
    """Register a callback to the event that the port's preamble insertion status changes."""


from typing import (
    TYPE_CHECKING,
    Type
)
from xoa_driver.internals.core.commands import (
    C_REMOTEPORTCOUNTS,
    C_BUILDSTRING,
)
from xoa_driver.internals.utils.modules_manager import ModulesManager
from xoa_driver.internals import revisions
from xoa_driver.internals import exceptions
if TYPE_CHECKING:
    from xoa_driver.internals.modules import module_l47 as ml47

from xoa_driver.internals.state_storage import testers_state
from ._base_tester import BaseTester
from .genuine import management_interface as mi


def get_module_type(revision: str) -> "Type":
    module_type = revisions.vulcan_modules.get(revision)
    if not module_type:
        raise exceptions.WrongModuleError(
            revision,
            set(revisions.vulcan_modules.keys()),
        )
    return module_type

class L47Tester(BaseTester["testers_state.GenuineTesterLocalState"]):
    """
    Representation of a physical Xena Vulcan Tester.
    """
    
    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, debug: bool = False ) -> None:
        super().__init__(host=host, username=username, password=password, port=port, debug=debug)
        
        self._local_states = testers_state.GenuineTesterLocalState(host, port)
        
        self.build_string = C_BUILDSTRING(self._conn)
        """
        Representation of C_BUILDSTRING
        """
        
        self.management_interface = mi.ManagementInterface(self._conn)
        """
        Tester management interface that includes IP address, DHCP, MAC address and hostname.
        """

        self.modules: ModulesManager["ml47.ModuleL47"] = ModulesManager(self._conn, get_module_type)
        """
        Module index manager of the tester.
        """
    
    @property
    def info(self) -> testers_state.GenuineTesterLocalState:
        return self._local_states
    
    async def _setup(self):
        await super()._setup()
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        
        ft_pc = await C_REMOTEPORTCOUNTS(self._conn).get()
        port_counts = ft_pc.port_counts
        await self.modules.fill_l47(port_counts)
        return self


import typing
from xoa_driver import ports
from xoa_driver.internals.utils import ports_manager as pm

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23

@typing.final
class MThor400G7S1P_b(ModuleL23):
    """Test module Thor-400G-7S-1P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PThor400G7S1P_b] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PThor400G7S1P_b, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Thor-400G-7S-1P[b]"""

@typing.final
class MThor400G7S1P_c(ModuleL23):
    """Test module Thor-400G-7S-1P[c]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PThor400G7S1P_c] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PThor400G7S1P_c, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Thor-400G-7S-1P[c]"""

@typing.final
class MFreya800G1S1P_a(ModuleL23):
    """Test module Freya-800G-1S-1P[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_a] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PFreya800G1S1P_a, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Freya-800G-1S-1P[a]"""

@typing.final
class MFreya800G1S1P_b(ModuleL23):
    """Test module Freya-800G-1S-1P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_b] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PFreya800G1S1P_b, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Freya-800G-1S-1P[b]"""

@typing.final
class MFreya800G1S1POSFP_a(ModuleL23):
    """Test module Freya-800G-1S-1P-OSFP[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_a] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PFreya800G1S1POSFP_a, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Freya-800G-1S-1P-OSFP[a]"""


@typing.final
class MFreya800G4S1P_a(ModuleL23):
    """Test module Freya-800G-4S-1P[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_a] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PFreya800G4S1P_a, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Freya-800G-4S-1P[a]"""

@typing.final
class MFreya800G4S1POSFP_a(ModuleL23):
    """Test module Freya-800G-4S-1P-OSFP[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_a] = pm.PortsManager(
            conn=conn, 
            ports_type=ports.PFreya800G4S1POSFP_a, 
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port index manager of Freya-800G-4S-1P-OSFP[a]"""


import functools
from typing import TYPE_CHECKING
from xoa_driver.internals.ports import base_port

from xoa_driver.internals.core.commands import (
    P4_TRAFFIC,
    P4_STATE,
    P4_CAPABILITIES,
    P4_STATE_STATUS,
    P4_VLAN_OFFLOAD,
    P4_ARP_CONFIG,
    P4_NDP_CONFIG,
    P4_CAPTURE,
    P4_CAPTURE_GET_FIRST,
    P4_CAPTURE_GET_NEXT,
    P4_TX_PACKET_SIZE,
    P4_RX_PACKET_SIZE,
    P4_TX_MTU,
    P4_RX_MTU,
    P4_CLEAR,
    P4_SPEEDSELECTION,
    P4_MAX_PACKET_RATE,
    P4_PCI_INFO,
    P4_FW_VER,
    P4_DEV_NAME,
    P4_PORT_TYPE,
    P4_LICENSE_INFO,
    P4_APTITUDES,
)

if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.state_storage import ports_state
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as im
from xoa_driver.internals.indices.connection_group.cg import ConnectionGroupIdx

from .counters import PCounters
from .packet_engine import PacketEngine

class PCapture:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.start = P4_CAPTURE(conn, module_id, port_id)
        self.get_first_frame = P4_CAPTURE_GET_FIRST(conn, module_id, port_id)
        self.get_next_frame = P4_CAPTURE_GET_NEXT(conn, module_id, port_id)



class PortL47(base_port.BasePort["ports_state.PortL47LocalState"]):
    """L47 Port"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        
        self._local_states = ports_state.PortL47LocalState()
        
        self.traffic = P4_TRAFFIC(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_TRAFFIC`"""
        self.clear = P4_CLEAR(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_CLEAR`"""
        self.state = P4_STATE(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_STATE`"""
        self.type = P4_PORT_TYPE(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_PORT_TYPE`"""
        self.aptitudes = P4_APTITUDES(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_APTITUDES`"""
        self.last_state_status = P4_STATE_STATUS(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_STATE_STATUS`"""
        
        self.nic_name = P4_DEV_NAME(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_DEV_NAME`"""
        self.nic_firmware_version = P4_FW_VER(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_FW_VER`"""
        self.vlan_offload = P4_VLAN_OFFLOAD(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_VLAN_OFFLOAD`"""
        
        self.max_packet_rate = P4_MAX_PACKET_RATE(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_MAX_PACKET_RATE`"""
        self.capabilities = P4_CAPABILITIES(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_CAPABILITIES`"""
        self.speed_selection = P4_SPEEDSELECTION(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_SPEEDSELECTION`"""
        
        # region Not sure if this commands must be atthis level
        self.tx_packet_size = P4_TX_PACKET_SIZE(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_TX_PACKET_SIZE`"""
        self.rx_packet_size = P4_RX_PACKET_SIZE(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_RX_PACKET_SIZE`"""
        self.tx_mtu = P4_TX_MTU(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_TX_MTU`"""
        self.rx_mtu = P4_RX_MTU(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_RX_MTU`"""
        # endregion
        
        self.pci_info = P4_PCI_INFO(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_PCI_INFO`"""
        self.license_info = P4_LICENSE_INFO(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_LICENSE_INFO`"""
        
        self.arp_config = P4_ARP_CONFIG(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_ARP_CONFIG`"""
        self.ndp_config = P4_NDP_CONFIG(self._conn, module_id, port_id)
        """Representation of :class:`~xoa_driver.internals.core.commands.p4_commands.P4_NDP_CONFIG`"""
        
        self.capture = PCapture(self._conn, module_id, port_id)
        """L47 packet capture"""
        self.counters = PCounters(self._conn, module_id, port_id)
        """L47 counters"""
        self.packet_engine = PacketEngine(self._conn, module_id, port_id)
        """L47 packet engine"""
        
        self.connection_groups: "im.IndexManager[ConnectionGroupIdx]" = im.IndexManager(
            self._conn, 
            ConnectionGroupIdx, 
            module_id, 
            port_id
        )
        """L47 connection group index manager."""
    
    @property
    def info(self) -> ports_state.PortL47LocalState:
        return self._local_states
    
    async def _setup(self):
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        return self
    
    on_capabilities_change = functools.partialmethod(utils.on_event, P4_CAPABILITIES)
    """Register a callback to the event that the L47 port's capabilities change."""
    on_speed_selection_change = functools.partialmethod(utils.on_event, P4_SPEEDSELECTION)
    """Register a callback to the event that the L47 port's speed mode changes."""
    on_state_change = functools.partialmethod(utils.on_event, P4_STATE)
    """Register a callback to the event that the L47 port's state changes."""
    on_license_info_change = functools.partialmethod(utils.on_event, P4_LICENSE_INFO)
    """Register a callback to the event that the L47 port's license information changes."""



from xoa_driver import modules

# region Valkyrie Modules

valkyrie_modules = {
    # family_d
    "Odin-1G-3S-6P-E": modules.MOdin1G3S6PE,
    "Odin-1G-3S-6P": modules.MOdin1G3S6P,
    "Odin-1G-3S-6P[b]": modules.MOdin1G3S6P_b,
    "Odin-1G-3S-2P-T": modules.MOdin1G3S2PT,
    
    # family_e
    "Odin-5G-4S-6P-CU": modules.MOdin5G4S6PCU,
    "Odin-10G-5S-6P-CU": modules.MOdin10G5S6PCU,
    "Odin-10G-5S-6P-CU[b]": modules.MOdin10G5S6PCU_b,
    "Odin-10G-3S-6P-CU": modules.MOdin10G3S6PCU,
    "Odin-10G-3S-2P-CU": modules.MOdin10G3S2PCU,
    
    # family_f
    "Odin-10G-1S-2P": modules.MOdin10G1S2P,
    "Odin-10G-1S-2P[b]": modules.MOdin10G1S2P_b,
    "Odin-10G-1S-2P[c]": modules.MOdin10G1S2P_c,
    "Odin-10G-1S-6P": modules.MOdin10G1S6P,
    "Odin-10G-1S-6P[b]": modules.MOdin10G1S6P_b,
    "Odin-10G-1S-2P-T": modules.MOdin10G1S2PT,
    "Odin-10G-1S-2P[d]": modules.MOdin10G1S2P_d,
    "Odin-10G-1S-12P": modules.MOdin10G1S12P,
    
    # family_g
    "Loki-100G-3S-1P": modules.MLoki100G3S1P,
    "Loki-100G-3S-1P[b]": modules.MLoki100G3S1P_b,
    "Loki-100G-3S-1P-SE": modules.MLoki100G3S1PSE,
    
    # family_h
    "Loki-100G-5S-1P": modules.MLoki100G5S1P,
    
    # family_i
    "Loki-100G-5S-2P": modules.MLoki100G5S2P,
    
    # family_j
    "Thor-100G-5S-4P": modules.MThor100G5S4P,
    
    # family_k
    "Thor-400G-7S-1P": modules.MThor400G7S1P,
    
    # family_l
    "Thor-400G-7S-1P[b]": modules.MThor400G7S1P_b,
    "Thor-400G-7S-1P[c]": modules.MThor400G7S1P_c,
    "Freya-800G-1S-1P[a]": modules.MFreya800G1S1P_a,
    "Freya-800G-1S-1P[b]": modules.MFreya800G1S1P_b,
    "Freya-800G-1S-1P-OSFP[a]": modules.MFreya800G1S1POSFP_a,
    "Freya-800G-4S-1[a]": modules.MFreya800G4S1P_a,
    "Freya-800G-4S-1P-OSFP[a]": modules.MFreya800G4S1POSFP_a,
    
    # family_m
    "Odin-1G-3S-6P-T1-RJ45": modules.MOdin1G3S6PT1RJ45,
    
    # family_n
    "Odin-40G-2S-2P": modules.MOdin40G2S2P,
    
    # family_combi
    "Odin-10G-4S-2P-Combi": modules.MOdin10G4S2PCombi,
    "Odin-10G-4S-2P-Combi[b]": modules.MOdin10G4S2PCombi_b,
    
}

# endregion

# region Chimera Modules

chimera_modules = {
    "Chimera-100G-5S-2P": modules.ModuleChimera,
}

# endregion

# region Vulcan Modules

vulcan_modules = {
    "Xena L47 Module": modules.ModuleL47,
}

# endregion



from typing import TypeVar
from xoa_driver.internals.core.commands.enums import FilterType

T = TypeVar("T")

def prevent_set(inst: T, filter_type: FilterType = FilterType.WORKING) -> T:
    METHOD_NAME = "set"
    def _set(*args, **kwargs): raise Exception(f"Method {METHOD_NAME} is not allowed")
    if filter_type is FilterType.WORKING and hasattr(inst, METHOD_NAME):
        setattr(inst, METHOD_NAME, _set)
    return inst


#: All exception classes which can be propagated to the upper level.

from .internals.exceptions import *
from .internals.core.transporter.exceptions import EstablishConnectionError, BadStatus

__all__ = (
    "WrongModuleError",
    "WrongTesterError",
    "WrongTesterPasswordError",
    "EstablishConnectionError",
    "BadStatus",
)   


from typing import (
    TYPE_CHECKING,
    TypeVar,
    Generic,
)
from dataclasses import dataclass
if TYPE_CHECKING:
    from .. import interfaces
    from .. import protocol


AwaitableDataType = TypeVar("AwaitableDataType") 
@dataclass(frozen=True)
class Token(Generic[AwaitableDataType]):
    """
    A wrapper of connection and request. 
    Which can be used for await an single command or organized in to the ordered 
    sequence of the commands which will be send to the server in single request.
    """
    __slots__ = ('connection', 'request')
    connection: "interfaces.IConnection"
    request: "protocol.Request"

    def __await__(self):
        return self.__ask().__await__()

    async def __ask(self) -> AwaitableDataType:
        (
            data, 
            fut
        ) = await self.connection.prepare_data(self.request)
        self.connection.send(data)
        try:
            result = await fut
        except Exception as e:
            raise e
        else:
            return result


import asyncio
import functools
from typing import TYPE_CHECKING
from xoa_driver.internals.core.commands import (
    M_STATUS,
    M_UPGRADE,
    M_UPGRADEPROGRESS,
    M_CFPTYPE,
    M_CFPCONFIG,
    M_COMMENT,
    M_CAPABILITIES,
    M_CLOCKPPB,
    M_TXCLOCKSOURCE_NEW,
    M_TXCLOCKSTATUS_NEW,
    M_EMULBYPASS,
    # M_LATENCYMODE, #TODO: wait for val-90 to assign the correct command code
)

from xoa_driver.internals.utils import ports_manager as pm
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.state_storage import modules_state
from xoa_driver import ports
from . import base_module as bm
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from . import __interfaces as m_itf

class ChTXClock:
    """
    Advanced timing feature (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.source = M_TXCLOCKSOURCE_NEW(conn, module_id)
        """
        The source that drives the TX clock rate of the ports on the test module.
        Representation of M_TXCLOCKSOURCE_NEW
        """
        self.status = M_TXCLOCKSTATUS_NEW(conn, module_id)
        """
        TX clock status of the test module.
        Representation of M_TXCLOCKSTATUS_NEW
        """


class ChCFP:
    """
    CFP test module (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.type = M_CFPTYPE(conn, module_id)
        """
        The transceiver's CFP type currently inserted.
        Representation of M_CFPTYPE
        """
        self.config = M_CFPCONFIG(conn, module_id)
        """
        The CFP configuration of the test module.
        Representation of M_CFPCONFIG
        """


class ChUpgrade:
    """
    Upgrade test module (Chimera).
    """
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.start = M_UPGRADE(conn, module_id)
        """
        Start the upgrade progress of the test module.
        Representation of M_UPGRADE
        """
        self.progress = M_UPGRADEPROGRESS(conn, module_id)
        """
        Upgrade progress status of the test module.
        Representation of M_UPGRADEPROGRESS
        """

class ModuleChimera(bm.BaseModule["modules_state.ModuleLocalState"]):
    """
    Representation of a Chimera module on physical tester.
    """
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        
        self._local_states = modules_state.ModuleLocalState()
        
        self.tx_clock = ChTXClock(conn, self.module_id)
        """
        Advanced timing feature (Chimera).
        """
        self.cfp = ChCFP(conn, self.module_id)
        """
        CFP test module (Chimera).
        """
        self.upgrade = ChUpgrade(conn, self.module_id)
        """
        Upgrade test module (Chimera).
        """

        self.capabilities = M_CAPABILITIES(conn, self.module_id)
        """
        Test module's capabilities.
        Representation of M_CAPABILITIES
        """

        self.comment = M_COMMENT(conn, self.module_id)
        """
        Test module's description.
        Representation of M_COMMENT
        """

        self.status = M_STATUS(conn, self.module_id)
        """
        Test module's status.
        Representation of M_STATUS
        """

        self.clock_ppb = M_CLOCKPPB(conn, self.module_id)
        """
        Test module's local clock adjustment.
        Representation of M_CLOCKPPB
        """

        self.emulator_bypass_mode = M_EMULBYPASS(conn, self.module_id)
        """
        Bypass mode of the Chimera module.
        Representation of M_EMULBYPASS
        """

        self.ports: pm.PortsManager["ports.PortChimera"] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PortChimera,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """
        Port index manager of the Chimera module.
        """
    
    @property
    def info(self) -> modules_state.ModuleLocalState:
        return self._local_states
    
    async def _setup(self):
        await asyncio.gather(
            self._local_states.initiate(self),
            self.ports.fill()
        )
        self._local_states.register_subscriptions(self)
        return self

    on_cfp_type_change = functools.partialmethod(utils.on_event, M_CFPTYPE)
    """
    Register a callback to the event that the module's CFP type changes.
    """

    on_cfp_config_change = functools.partialmethod(utils.on_event, M_CFPCONFIG)
    """
    Register a callback to the event that the module's CFP configuration changes.
    """

    on_model_change = functools.partialmethod(utils.on_event, M_STATUS)
    """
    Register a callback to the event that the module's model changes.
    """

    # on_latency_mode_change = functools.partialmethod(utils.on_event, M_LATENCYMODE)



import functools
from typing import TYPE_CHECKING
from .bases.port_l23_genuine import BasePortL23Genuine
from xoa_driver.internals.core.commands import (
    P_AUTONEGSELECTION, # questinable which ports are electrical
    P_LPENABLE,
    P_LPTXMODE,
    P_LPSTATUS,
    P_LPPARTNERAUTONEG,
    P_LPSNRMARGIN,
    P_LPRXPOWER,
)
from xoa_driver.internals.utils import attributes as utils
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf


class LowPowerMode:
    """L23 port low power mode."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.enable = P_LPENABLE(conn, module_id, port_id)
        """Energy Efficiency Ethernet.
        Representation of P_LPENABLE
        """
        
        self.mode = P_LPTXMODE(conn, module_id, port_id)
        """Low power TX mode.
        Representation of P_LPTXMODE
        """
        self.status = P_LPSTATUS(conn, module_id, port_id)
        """Low power status.
        Representation of P_LPSTATUS
        """

        self.partner_autonegotiation = P_LPPARTNERAUTONEG(conn, module_id, port_id)
        """EEE capabilities advertised during autonegotiation by the far side.
        Representation of P_LPPARTNERAUTONEG
        """

        self.snr_margin = P_LPSNRMARGIN(conn, module_id, port_id)
        """SNR margin on the four link channels by the PHY.
        Representation of P_LPSNRMARGIN
        """

        self.rx_power = P_LPRXPOWER(conn, module_id, port_id)
        """RX power recorded during training for the four channels.
        Representation of P_LPRXPOWER
        """


class FamilyE(BasePortL23Genuine):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.autoneg_selection = P_AUTONEGSELECTION(conn, module_id, port_id)
        """L23 port's auto-negotiation selection.
        Representation of P_AUTONEGSELECTION
        """

        self.eee = LowPowerMode(conn, module_id, port_id)
        """L23 port Low Power mode settings."""


    on_autoneg_selection_change = functools.partialmethod(utils.on_event, P_AUTONEGSELECTION)
    """Register a callback to the event that the L23 port's auto-negotiation selection changes."""

    on_low_power_mode_enable_change = functools.partialmethod(utils.on_event, P_LPENABLE)
    """Register a callback to the event that the L23 port's low power status changes."""

    on_low_power_mode_change = functools.partialmethod(utils.on_event, P_LPTXMODE)
    """Register a callback to the event that the L23 port's low power mode changes."""



class POdin5G4S6PCU(FamilyE):
    """L23 port on Odin-5G-4S-6P-CU module.
    """
    ...

class POdin10G5S6PCU(FamilyE):
    """L23 port on Odin-10G-5S-6P-CU module.
    """
    ...

class POdin10G5S6PCU_b(FamilyE):
    """L23 port on Odin-10G-5S-6P-CU[b] module.
    """
    ...

class POdin10G3S6PCU(FamilyE):
    """L23 port on Odin-10G-3S-6P-CU module.
    """
    ...

class POdin10G3S2PCU(FamilyE):
    """L23 port on Odin-10G-3S-2P-CU module.
    """
    ...




from dataclasses import dataclass
from ..protocol.fields.data_types import *
from ..protocol.fields.add_on import add_on
from ..protocol.fields.field import XmpField
from .enums import OnOff

@add_on
@dataclass
class ArpChunk:
    ipv4_address: XmpField[XmpIPV4Address] = XmpField(XmpIPV4Address)  # The IP address to match to the Target IP address in the ARP requests
    prefix: XmpField[XmpShort] = XmpField(XmpShort)  # The IP prefix used for address matching.
    patched_mac: XmpField[XmpByte] = XmpField(XmpByte, choices=OnOff)  # If on, the target MAC address will be patched with the part of the IP address that is not masked by the prefix
    mac_address: XmpField[XmpMacAddress] = XmpField(XmpMacAddress)  # The target MAC address to return in the ARP reply. If set t to all-zero, the port MAC address will be used.


class ArpChunkList(XmpDefaultList):
    element_type = ArpChunk # type: ignore


@add_on
@dataclass
class NdpChunk:
    ipv6_address: XmpField[XmpIPV6Address] = XmpField(XmpIPV6Address)  # The IP address to match to the Target IP address in the NDP Neighbor Solicitation
    prefix: XmpField[XmpShort] = XmpField(XmpShort)  # The IP prefix used for address matching.
    patched_mac: XmpField[XmpByte] = XmpField(XmpByte, choices=OnOff)  # If on, the target MAC address will be patched with the part of the IP address that is not masked by the prefix
    mac_address: XmpField[XmpMacAddress] = XmpField(XmpMacAddress)  # The target MAC address to return in the NDP Neighbor Advertisement. If set t to all-zero, the port MAC address will be used.


class NdpChunkList(XmpDefaultList):
    element_type = NdpChunk # type: ignore


@add_on
@dataclass
class PortSpeedChuck:
    port_count: XmpField[XmpInt] = XmpField(XmpInt)   # Number of ports for this speed.
    port_speed: XmpField[XmpInt] = XmpField(XmpInt)  # Unit is Mbps.

class PortSpeedChuckList(XmpDefaultList):
    element_type = PortSpeedChuck # type: ignore



